# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pytest
import json
import warnings
import numpy as np
import biotite.structure as struc
import biotite.structure.io as strucio
from biotite.structure.info import residue
from biotite.structure.residues import get_residue_masks
from biotite.structure.hbond import hbond
from os.path import join
from ..util import data_dir
# For ``base_pairs_edge()`` differences to a reference can be ambiguous
# as the number hydrogen bonds between two different edges can be equal.
# In order to distinguish ambiguously identified edges from wrongfully
# identified edges the full edge matrix, listing the number of hydrogen
# bonds for each edge has to be considered.
from biotite.structure.basepairs import _get_edge_matrix


def reversed_iterator(iter):
    """
    Returns a reversed list of the elements of an Iterator.
    """
    return reversed(list(iter))


@pytest.fixture
def nuc_sample_array():
    return strucio.load_structure(
        join(data_dir("structure"), "base_pairs", "1qxb.cif")
    )


@pytest.fixture
def basepairs(nuc_sample_array):
    """
    Generate a test output for the base_pairs function.
    """
    residue_indices, residue_names = struc.residues.get_residues(
        nuc_sample_array
    )[0:24]
    return np.vstack((residue_indices[:12], np.flip(residue_indices)[:12])).T

def check_residue_starts(computed_starts, nuc_sample_array):
    """
    Assert that computed starts are residue starts.
    """
    residue_starts = struc.get_residue_starts(nuc_sample_array)
    for start in computed_starts.flatten():
        assert start in residue_starts

def check_output(computed_basepairs, basepairs):
    """
    Check the output of base_pairs.
    """

    # Check if base pairs are unique in computed_basepairs
    seen = set()
    assert (not any(
        (base1, base2) in seen) or (base2, base1 in seen)
        or seen.add((base1, base2)) for base1, base2 in computed_basepairs
        )
    # Check if the right number of base pairs is in computed_base pairs
    assert(len(computed_basepairs) == len(basepairs))
    # Check if the right base pairs are in computed_basepairs
    for comp_basepair in computed_basepairs:
        assert ((comp_basepair in basepairs) \
                or (comp_basepair in np.flip(basepairs)))


@pytest.mark.parametrize("unique_bool", [False, True])
def test_base_pairs_forward(nuc_sample_array, basepairs, unique_bool):
    """
    Test for the function base_pairs.
    """
    computed_basepairs = struc.base_pairs(nuc_sample_array, unique=unique_bool)
    check_residue_starts(computed_basepairs, nuc_sample_array)
    check_output(nuc_sample_array[computed_basepairs].res_id, basepairs)


def test_base_pairs_forward_no_hydrogen(nuc_sample_array, basepairs):
    """
    Test for the function base_pairs with the hydrogens removed from the
    test structure.
    """
    nuc_sample_array = nuc_sample_array[nuc_sample_array.element != "H"]
    computed_basepairs = struc.base_pairs(nuc_sample_array)
    check_residue_starts(computed_basepairs, nuc_sample_array)
    check_output(nuc_sample_array[computed_basepairs].res_id, basepairs)


@pytest.mark.parametrize("unique_bool", [False, True])
def test_base_pairs_reverse(nuc_sample_array, basepairs, unique_bool):
    """
    Reverse the order of residues in the atom_array and then test the
    function base_pairs.
    """

    # Reverse sequence of residues in nuc_sample_array
    reversed_nuc_sample_array = struc.AtomArray(0)
    for residue in reversed_iterator(struc.residue_iter(nuc_sample_array)):
        reversed_nuc_sample_array = reversed_nuc_sample_array + residue

    computed_basepairs = struc.base_pairs(
        reversed_nuc_sample_array, unique=unique_bool
    )
    check_residue_starts(computed_basepairs, reversed_nuc_sample_array)
    check_output(
        reversed_nuc_sample_array[computed_basepairs].res_id, basepairs
    )


def test_base_pairs_reverse_no_hydrogen(nuc_sample_array, basepairs):
    """
    Remove the hydrogens from the sample structure. Then reverse the
    order of residues in the atom_array and then test the function
    base_pairs.
    """
    nuc_sample_array = nuc_sample_array[nuc_sample_array.element != "H"]
    # Reverse sequence of residues in nuc_sample_array
    reversed_nuc_sample_array = struc.AtomArray(0)
    for residue in reversed_iterator(struc.residue_iter(nuc_sample_array)):
        reversed_nuc_sample_array = reversed_nuc_sample_array + residue

    computed_basepairs = struc.base_pairs(reversed_nuc_sample_array)
    check_residue_starts(computed_basepairs, reversed_nuc_sample_array)
    check_output(
        reversed_nuc_sample_array[computed_basepairs].res_id, basepairs
    )

def test_base_pairs_incomplete_structure(nuc_sample_array):
    """
    Remove atoms belonging to the pyrimidine / purine ring of each base
    and the ``O2`` atom contained in pyrimidine bases.

    Test that no base pairs are detected as all bases have less than 3 
    common atoms with their implemented reference base. 
    """
    
    nuc_sample_array = nuc_sample_array[
        ~ np.isin(
            nuc_sample_array.atom_name, 
            ['N1', 'C2', 'N3', 'C4', 'C5', 'C6', 'N7', 'C8', 'N9', 'O2']
        )
    ]
    with pytest.warns(struc.IncompleteStructureWarning):
        assert len(struc.base_pairs(nuc_sample_array)) == 0

@pytest.mark.parametrize("seed", range(10))
def test_base_pairs_reordered(nuc_sample_array, seed):
    """
    Test the function base_pairs with structure where the atoms are not
    in the RCSB-Order.
    """
    # Randomly reorder the atoms in each residue
    nuc_sample_array_reordered = struc.AtomArray(0)
    np.random.seed(seed)

    for residue in struc.residue_iter(nuc_sample_array):
        bound = residue.array_length()
        indices = np.random.choice(
            np.arange(bound), bound,replace=False
        )
        nuc_sample_array_reordered += residue[..., indices]

    assert(np.all(
        struc.base_pairs(nuc_sample_array)
        == struc.base_pairs(nuc_sample_array_reordered)
    ))


def test_map_nucleotide():
    """Test the function map_nucleotide with some examples.
    """
    pyrimidines = ['C', 'T', 'U']
    purines = ['A', 'G']

    # Test that the standard bases are correctly identified
    assert struc.map_nucleotide(residue('U')) == ('U', True)
    assert struc.map_nucleotide(residue('A')) == ('A', True)
    assert struc.map_nucleotide(residue('T')) == ('T', True)
    assert struc.map_nucleotide(residue('G')) == ('G', True)
    assert struc.map_nucleotide(residue('C')) == ('C', True)

    # Test that some non_standard nucleotides are mapped correctly to
    # pyrimidine/purine references
    psu_tuple = struc.map_nucleotide(residue('PSU'))
    assert psu_tuple[0] in pyrimidines
    assert psu_tuple[1] == False

    psu_tuple = struc.map_nucleotide(residue('3MC'))
    assert psu_tuple[0] in pyrimidines
    assert psu_tuple[1] == False

    i_tuple = struc.map_nucleotide(residue('I'))
    assert i_tuple[0] in purines
    assert i_tuple[1] == False

    m7g_tuple = struc.map_nucleotide(residue('M7G'))
    assert m7g_tuple[0] in purines
    assert m7g_tuple[1] == False

    with pytest.warns(struc.IncompleteStructureWarning):
        assert struc.map_nucleotide(residue('ALA')) == (None, False)


def get_reference(pdb_id, suffix):
    """
    Gets a reference structure and a related json array depending on
    a specified JSON-suffix and PDB ID.
    """
    structure = strucio.load_structure(
        join(data_dir("structure"), "base_pairs", f"{pdb_id}.cif")
    )

    with open(
        join(data_dir("structure"), "base_pairs", f"{pdb_id}_{suffix}.json"
    ), "r") as file:
        reference = np.array(json.load(file))

    return structure, reference

def get_reference_index(pair, array):
    """
    Get the index of the row in a reference array, where the first two
    columns match the values of ``pair``. If no match is found ``Ǹone``
    is returned.
    """
    pair = sorted(pair)
    if np.any((array[:, 0] == pair[0]) & (array[:, 1] == pair[1])):
        return np.where((array[:, 0] == pair[0]) & (array[:, 1] == pair[1]))
    return None



def check_edge_plausibility(
    reference_structure, pair, reference_edges, output_edges
):
    """
    Checks if the difference to a reference edge is at least ambiguous.
    A difference is defined as ambiguous, if the number of hydrogen
    bonds on a given edge is the same as on the edge specified in the
    reference.
    """
    # Get the complete edge matrix for a given edge
    edge_matrix = _get_edge_matrix(
        reference_structure, get_residue_masks(reference_structure, pair)
    )
    # Check if the difference to the reference is at least arbitrary
    for edges, reference_edge, output_edge in zip(
        edge_matrix, reference_edges, output_edges
    ):
        max_matches = np.max(edges)
        # The edge type corresponds to the index in the edge matrix + 1
        max_match_edges = np.argwhere(edges == max_matches).flatten() + 1
        assert reference_edge in max_match_edges
        assert output_edge in max_match_edges


@pytest.mark.parametrize("pdb_id", ["1gid", "1nkw", "1xnr"])
def test_base_pairs_edge(pdb_id):
    """
    Test the function ``base_pairs_edge``. Each test structure is a
    crystal structure onto which hydrogens were added using Gromacs
    force fields. The reference data was taken from the NDB-database
    annotations and parsed as json array.
    """
    # Get the references
    reference_structure, reference_edges = get_reference(pdb_id, "edges")
    # Calculate base pairs and edges for the references
    pairs = struc.base_pairs(reference_structure)
    edges = struc.base_pairs_edge(reference_structure, pairs)

    # Check the plausibility with the reference data for each base pair
    for pair, pair_edges in zip(pairs, edges):
        pair_res_ids = reference_structure[pair].res_id
        index = get_reference_index(pair_res_ids, reference_edges)
        if index is not None:
            pair_reference_edges =  [
                reference_edges[index, 2], reference_edges[index, 3]
            ]
            check_edge_plausibility(
                reference_structure, pair, pair_reference_edges, pair_edges
            )


@pytest.mark.parametrize("pdb_id", ["1gid", "1nkw", "1xnr"])
def test_base_pairs_glycosidic_bond(pdb_id):
    """
    Test the function ``base_pairs_edge``. Each test structure is a
    crystal structure onto which hydrogens were added using Gromacs
    force fields. The reference data was taken from the NDB-database
    annotations and parsed as json array.
    """
    # Get the references
    reference_structure, reference_gly_bonds = get_reference(pdb_id, "sugar")
    # Calculate base pairs and edges for the references
    pairs = struc.base_pairs(reference_structure)
    glycosidic_bond_orientations = struc.base_pairs_glycosidic_bond(
        reference_structure, pairs
    )

    # Check the plausibility with the reference data for each base pair
    for pair, pair_orientation in zip(pairs, glycosidic_bond_orientations):
        pair_res_ids = reference_structure[pair].res_id
        index = get_reference_index(pair_res_ids, reference_gly_bonds)
        if index is not None:
            reference_orientation = struc.GlycosidicBond(
                reference_gly_bonds[index, 2]
            )
            assert reference_orientation == pair_orientation


def test_base_stacking():
    """
    Test ``base_stacking()`` using the DNA-double-helix 1BNA. It is
    expected that adjacent bases are stacked. However, due to
    distortions in the helix there are exception for this particular
    helix.
    """
    # Load the test structure (1BNA) - a DNA-double-helix
    helix = strucio.load_structure(
        join(data_dir("structure"), "base_pairs", "1bna.mmtf")
    )

    residue_starts = struc.get_residue_starts(helix)

    # For a DNA-double-helix it is expected that adjacent bases are
    # stacked.
    expected_stackings = []
    for i in range(1, 24):
        expected_stackings.append([i, i+1])

    # Due to distortions in the helix not all adjacent bases have a
    # geometry that meets the criteria of `base_stacking`.
    expected_stackings.remove([10, 11])
    expected_stackings.remove([12, 13])
    expected_stackings.remove([13, 14])

    stacking = struc.base_stacking(helix)

    # Assert stacking outputs correct residue starts
    for stacking_start in stacking.flatten():
        assert stacking_start in residue_starts

    # Assert the number of stacking interactions is corrrect
    assert len(struc.base_stacking(helix)) == len(expected_stackings)

    # Assert the stacking interactions are correct
    for interaction in helix[stacking].res_id:
        assert list(interaction) in expected_stackings





# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides utility for separating structures into single
molecules.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["get_molecule_indices", "get_molecule_masks", "molecule_iter"]

import numpy as np
from .atoms import AtomArray, AtomArrayStack
from .bonds import BondList, find_connected


def get_molecule_indices(array):
    """
    Get an index array for each molecule in the given structure.

    A molecule is defined as a group of atoms that are directly or
    indirectly connected via covalent bonds.
    In this function a single atom, that has no connection to any other
    atom (e.g. an ion), also qualifies as a molecule.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack or BondList
        The input structure with an associated :class:`BondList`.
        Alternatively, the :class:`BondList` can be directly supplied.
    
    Returns
    -------
    indices : list of ndarray, dtype=int
        Each element in the list is an index array referring to the
        atoms of a single molecule.
        Consequently, the length of this list is equal to the number of
        molecules in the input `array`.
    
    See also
    --------
    get_molecule_masks
    molecule_iter

    Examples
    --------
    Get an :class:`AtomArray` for ATP and show that it is a single
    molecule:

    >>> atp = residue("ATP")
    >>> indices = get_molecule_indices(atp)
    >>> print(len(indices))
    1

    Separate ATP into two molecules by breaking the glycosidic bond
    to the triphosphate:

    >>> i, j = np.where(np.isin(atp.atom_name, ("O5'", "PA")))[0]
    >>> atp.bonds.remove_bond(i, j)
    >>> indices = get_molecule_indices(atp)
    >>> print(len(indices))
    2
    >>> print(atp[indices[0]])
    HET         0  ATP PG     P         1.200   -0.226   -6.850
    HET         0  ATP O1G    O         1.740    1.140   -6.672
    HET         0  ATP O2G    O         2.123   -1.036   -7.891
    HET         0  ATP O3G    O        -0.302   -0.139   -7.421
    HET         0  ATP PB     P         0.255   -0.130   -4.446
    HET         0  ATP O1B    O         0.810    1.234   -4.304
    HET         0  ATP O2B    O        -1.231   -0.044   -5.057
    HET         0  ATP O3B    O         1.192   -0.990   -5.433
    HET         0  ATP PA     P        -0.745    0.068   -2.071
    HET         0  ATP O1A    O        -2.097    0.143   -2.669
    HET         0  ATP O2A    O        -0.125    1.549   -1.957
    HET         0  ATP O3A    O         0.203   -0.840   -3.002
    HET         0  ATP HOG2   H         2.100   -0.546   -8.725
    HET         0  ATP HOG3   H        -0.616   -1.048   -7.522
    HET         0  ATP HOB2   H        -1.554   -0.952   -5.132
    HET         0  ATP HOA2   H         0.752    1.455   -1.563
    >>> print(atp[indices[1]])
    HET         0  ATP O5'    O        -0.844   -0.587   -0.604
    HET         0  ATP C5'    C        -1.694    0.260    0.170
    HET         0  ATP C4'    C        -1.831   -0.309    1.584
    HET         0  ATP O4'    O        -0.542   -0.355    2.234
    HET         0  ATP C3'    C        -2.683    0.630    2.465
    HET         0  ATP O3'    O        -4.033    0.165    2.534
    HET         0  ATP C2'    C        -2.011    0.555    3.856
    HET         0  ATP O2'    O        -2.926    0.043    4.827
    HET         0  ATP C1'    C        -0.830   -0.418    3.647
    HET         0  ATP N9     N         0.332    0.015    4.425
    HET         0  ATP C8     C         1.302    0.879    4.012
    HET         0  ATP N7     N         2.184    1.042    4.955
    HET         0  ATP C5     C         1.833    0.300    6.033
    HET         0  ATP C6     C         2.391    0.077    7.303
    HET         0  ATP N6     N         3.564    0.706    7.681
    HET         0  ATP N1     N         1.763   -0.747    8.135
    HET         0  ATP C2     C         0.644   -1.352    7.783
    HET         0  ATP N3     N         0.088   -1.178    6.602
    HET         0  ATP C4     C         0.644   -0.371    5.704
    HET         0  ATP H5'1   H        -2.678    0.312   -0.296
    HET         0  ATP H5'2   H        -1.263    1.259    0.221
    HET         0  ATP H4'    H        -2.275   -1.304    1.550
    HET         0  ATP H3'    H        -2.651    1.649    2.078
    HET         0  ATP HO3'   H        -4.515    0.788    3.094
    HET         0  ATP H2'    H        -1.646    1.537    4.157
    HET         0  ATP HO2'   H        -3.667    0.662    4.867
    HET         0  ATP H1'    H        -1.119   -1.430    3.931
    HET         0  ATP H8     H         1.334    1.357    3.044
    HET         0  ATP HN61   H         3.938    0.548    8.562
    HET         0  ATP HN62   H         4.015    1.303    7.064
    HET         0  ATP H2     H         0.166   -2.014    8.490
    """
    if isinstance(array, BondList):
        bonds = array
    elif isinstance(array, (AtomArray, AtomArrayStack)):
        if array.bonds is None:
            raise ValueError("An associated BondList is required")
        bonds = array.bonds
    else:
        raise TypeError(
            f"Expected a 'BondList', 'AtomArray' or 'AtomArrayStack', "
            f"not '{type(array).__name__}'"
        )
    
    molecule_indices = []
    visited_mask = np.zeros(bonds.get_atom_count(), dtype=bool)
    while not visited_mask.all():
        root = np.argmin(visited_mask)
        connected = find_connected(bonds, root)
        visited_mask[connected] = True
        molecule_indices.append(connected)
    return molecule_indices


def get_molecule_masks(array):
    """
    Get a boolean mask for each molecule in the given structure.

    A molecule is defined as a group of atoms that are directly or
    indirectly connected via covalent bonds.
    In this function a single atom, that has no connection to any other
    atom (e.g. an ion), also qualifies as a molecule.

    Parameters
    ----------
    array : AtomArray, shape=(n,) or AtomArrayStack, shape=(m,n) or BondList
        The input structure with an associated :class:`BondList`.
        Alternatively, the :class:`BondList` can be directly supplied.
    
    Returns
    -------
    masks : ndarray, shape=(k,n), dtype=bool, 
        Each element in the array is a boolean mask referring to the
        atoms of a single molecule.
        Consequently, the length of this list is equal to the number of
        molecules in the input `array`.
    
    See also
    --------
    get_molecule_indices
    molecule_iter

    Examples
    --------
    Get an :class:`AtomArray` for ATP and show that it is a single
    molecule:

    >>> atp = residue("ATP")
    >>> masks = get_molecule_masks(atp)
    >>> print(len(masks))
    1

    Separate ATP into two molecules by breaking the glycosidic bond
    to the triphosphate:

    >>> i, j = np.where(np.isin(atp.atom_name, ("O5'", "PA")))[0]
    >>> atp.bonds.remove_bond(i, j)
    >>> masks = get_molecule_masks(atp)
    >>> print(len(masks))
    2
    >>> print(atp[masks[0]])
    HET         0  ATP PG     P         1.200   -0.226   -6.850
    HET         0  ATP O1G    O         1.740    1.140   -6.672
    HET         0  ATP O2G    O         2.123   -1.036   -7.891
    HET         0  ATP O3G    O        -0.302   -0.139   -7.421
    HET         0  ATP PB     P         0.255   -0.130   -4.446
    HET         0  ATP O1B    O         0.810    1.234   -4.304
    HET         0  ATP O2B    O        -1.231   -0.044   -5.057
    HET         0  ATP O3B    O         1.192   -0.990   -5.433
    HET         0  ATP PA     P        -0.745    0.068   -2.071
    HET         0  ATP O1A    O        -2.097    0.143   -2.669
    HET         0  ATP O2A    O        -0.125    1.549   -1.957
    HET         0  ATP O3A    O         0.203   -0.840   -3.002
    HET         0  ATP HOG2   H         2.100   -0.546   -8.725
    HET         0  ATP HOG3   H        -0.616   -1.048   -7.522
    HET         0  ATP HOB2   H        -1.554   -0.952   -5.132
    HET         0  ATP HOA2   H         0.752    1.455   -1.563
    >>> print(atp[masks[1]])
    HET         0  ATP O5'    O        -0.844   -0.587   -0.604
    HET         0  ATP C5'    C        -1.694    0.260    0.170
    HET         0  ATP C4'    C        -1.831   -0.309    1.584
    HET         0  ATP O4'    O        -0.542   -0.355    2.234
    HET         0  ATP C3'    C        -2.683    0.630    2.465
    HET         0  ATP O3'    O        -4.033    0.165    2.534
    HET         0  ATP C2'    C        -2.011    0.555    3.856
    HET         0  ATP O2'    O        -2.926    0.043    4.827
    HET         0  ATP C1'    C        -0.830   -0.418    3.647
    HET         0  ATP N9     N         0.332    0.015    4.425
    HET         0  ATP C8     C         1.302    0.879    4.012
    HET         0  ATP N7     N         2.184    1.042    4.955
    HET         0  ATP C5     C         1.833    0.300    6.033
    HET         0  ATP C6     C         2.391    0.077    7.303
    HET         0  ATP N6     N         3.564    0.706    7.681
    HET         0  ATP N1     N         1.763   -0.747    8.135
    HET         0  ATP C2     C         0.644   -1.352    7.783
    HET         0  ATP N3     N         0.088   -1.178    6.602
    HET         0  ATP C4     C         0.644   -0.371    5.704
    HET         0  ATP H5'1   H        -2.678    0.312   -0.296
    HET         0  ATP H5'2   H        -1.263    1.259    0.221
    HET         0  ATP H4'    H        -2.275   -1.304    1.550
    HET         0  ATP H3'    H        -2.651    1.649    2.078
    HET         0  ATP HO3'   H        -4.515    0.788    3.094
    HET         0  ATP H2'    H        -1.646    1.537    4.157
    HET         0  ATP HO2'   H        -3.667    0.662    4.867
    HET         0  ATP H1'    H        -1.119   -1.430    3.931
    HET         0  ATP H8     H         1.334    1.357    3.044
    HET         0  ATP HN61   H         3.938    0.548    8.562
    HET         0  ATP HN62   H         4.015    1.303    7.064
    HET         0  ATP H2     H         0.166   -2.014    8.490
    """
    if isinstance(array, BondList):
        bonds = array
    elif isinstance(array, (AtomArray, AtomArrayStack)):
        if array.bonds is None:
            raise ValueError("An associated BondList is required")
        bonds = array.bonds
    else:
        raise TypeError(
            f"Expected a 'BondList', 'AtomArray' or 'AtomArrayStack', "
            f"not '{type(array).__name__}'"
        )
    
    molecule_indices = get_molecule_indices(bonds)
    molecule_masks = np.zeros(
        (len(molecule_indices), bonds.get_atom_count()),
        dtype=bool
    )
    for i in range(len(molecule_indices)):
        molecule_masks[i, molecule_indices[i]] = True
    return molecule_masks


def molecule_iter(array):
    """
    Iterate over each molecule in a input structure.

    A molecule is defined as a group of atoms that are directly or
    indirectly connected via covalent bonds.
    In this function a single atom, that has no connection to any other
    atom (e.g. an ion), also qualifies as a molecule.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The input structure with an associated :class:`BondList`.
    
    Yields
    ------
    molecule : AtomArray or AtomArrayStack
        A single molecule of the input `array`.
    
    See also
    --------
    get_molecule_indices
    get_molecule_masks

    Examples
    --------
    Get an :class:`AtomArray` for ATP and break it into two molecules
    at the glycosidic bond to the triphosphate:

    >>> atp = residue("ATP")
    >>> i, j = np.where(np.isin(atp.atom_name, ("O5'", "PA")))[0]
    >>> atp.bonds.remove_bond(i, j)
    >>> for molecule in molecule_iter(atp):
    ...     print("New molecule")
    ...     print(molecule)
    ...     print()
    New molecule
    HET         0  ATP PG     P         1.200   -0.226   -6.850
    HET         0  ATP O1G    O         1.740    1.140   -6.672
    HET         0  ATP O2G    O         2.123   -1.036   -7.891
    HET         0  ATP O3G    O        -0.302   -0.139   -7.421
    HET         0  ATP PB     P         0.255   -0.130   -4.446
    HET         0  ATP O1B    O         0.810    1.234   -4.304
    HET         0  ATP O2B    O        -1.231   -0.044   -5.057
    HET         0  ATP O3B    O         1.192   -0.990   -5.433
    HET         0  ATP PA     P        -0.745    0.068   -2.071
    HET         0  ATP O1A    O        -2.097    0.143   -2.669
    HET         0  ATP O2A    O        -0.125    1.549   -1.957
    HET         0  ATP O3A    O         0.203   -0.840   -3.002
    HET         0  ATP HOG2   H         2.100   -0.546   -8.725
    HET         0  ATP HOG3   H        -0.616   -1.048   -7.522
    HET         0  ATP HOB2   H        -1.554   -0.952   -5.132
    HET         0  ATP HOA2   H         0.752    1.455   -1.563
    <BLANKLINE>
    New molecule
    HET         0  ATP O5'    O        -0.844   -0.587   -0.604
    HET         0  ATP C5'    C        -1.694    0.260    0.170
    HET         0  ATP C4'    C        -1.831   -0.309    1.584
    HET         0  ATP O4'    O        -0.542   -0.355    2.234
    HET         0  ATP C3'    C        -2.683    0.630    2.465
    HET         0  ATP O3'    O        -4.033    0.165    2.534
    HET         0  ATP C2'    C        -2.011    0.555    3.856
    HET         0  ATP O2'    O        -2.926    0.043    4.827
    HET         0  ATP C1'    C        -0.830   -0.418    3.647
    HET         0  ATP N9     N         0.332    0.015    4.425
    HET         0  ATP C8     C         1.302    0.879    4.012
    HET         0  ATP N7     N         2.184    1.042    4.955
    HET         0  ATP C5     C         1.833    0.300    6.033
    HET         0  ATP C6     C         2.391    0.077    7.303
    HET         0  ATP N6     N         3.564    0.706    7.681
    HET         0  ATP N1     N         1.763   -0.747    8.135
    HET         0  ATP C2     C         0.644   -1.352    7.783
    HET         0  ATP N3     N         0.088   -1.178    6.602
    HET         0  ATP C4     C         0.644   -0.371    5.704
    HET         0  ATP H5'1   H        -2.678    0.312   -0.296
    HET         0  ATP H5'2   H        -1.263    1.259    0.221
    HET         0  ATP H4'    H        -2.275   -1.304    1.550
    HET         0  ATP H3'    H        -2.651    1.649    2.078
    HET         0  ATP HO3'   H        -4.515    0.788    3.094
    HET         0  ATP H2'    H        -1.646    1.537    4.157
    HET         0  ATP HO2'   H        -3.667    0.662    4.867
    HET         0  ATP H1'    H        -1.119   -1.430    3.931
    HET         0  ATP H8     H         1.334    1.357    3.044
    HET         0  ATP HN61   H         3.938    0.548    8.562
    HET         0  ATP HN62   H         4.015    1.303    7.064
    HET         0  ATP H2     H         0.166   -2.014    8.490
    <BLANKLINE>
    """
    if array.bonds is None:
        raise ValueError("An associated BondList is required")
    bonds = array.bonds
    
    visited_mask = np.zeros(bonds.get_atom_count(), dtype=bool)
    while not visited_mask.all():
        # Take the first atom that has not been considered yet as root
        root = np.argmin(visited_mask)
        connected = find_connected(bonds, root)
        visited_mask[connected] = True
        yield array[..., connected]



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module allows checking of atom arrays and atom array stacks for
errors in the structure.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann, Daniel Bauer"
__all__ = ["check_id_continuity", "check_atom_id_continuity",
           "check_res_id_continuity", "check_bond_continuity",
           "check_duplicate_atoms",
           "renumber_atom_ids", "renumber_res_ids"]

import numpy as np
import warnings
from .atoms import Atom, AtomArray, AtomArrayStack
from .filter import filter_backbone
from .box import coord_to_fraction


def _check_continuity(array):
    diff = np.diff(array)
    discontinuity = np.where( ((diff != 0) & (diff != 1)) )
    return discontinuity[0] + 1


def check_id_continuity(array):
    """
    Check if the residue IDs are incremented by more than 1 or
    decremented, from one atom to the next one.
    
    An increment by more than 1 is as strong clue for missing residues,
    a decrement means probably a start of a new chain.

    DEPRECATED: Use :func:`check_res_id_continuity()` instead.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be checked.
    
    Returns
    -------
    discontinuity : ndarray, dtype=bool
        Contains the indices of atoms after a discontinuity
    """
    warnings.warn(
        "'check_id_continuity()' is deprecated, "
        "use 'check_res_id_continuity()' instead",
        DeprecationWarning
    )
    return check_res_id_continuity(array)


def check_atom_id_continuity(array):
    """
    Check if the atom IDs are incremented by more than 1 or
    decremented, from one atom to the next one.
    
    An increment by more than 1 is as strong clue for missing atoms.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be checked.
    
    Returns
    -------
    discontinuity : ndarray, dtype=bool
        Contains the indices of atoms after a discontinuity
    """
    ids = array.atom_id
    return _check_continuity(ids)


def check_res_id_continuity(array):
    """
    Check if the residue IDs are incremented by more than 1 or
    decremented, from one atom to the next one.
    
    An increment by more than 1 is as strong clue for missing residues,
    a decrement means probably a start of a new chain.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be checked.
    
    Returns
    -------
    discontinuity : ndarray, dtype=bool
        Contains the indices of atoms after a discontinuity
    """
    ids = array.res_id
    return _check_continuity(ids)


def check_bond_continuity(array, min_len=1.2, max_len=1.8):
    """
    Check if the peptide backbone atoms ("N","CA","C") have a
    non-reasonable distance to the next atom.
    
    A large or very small distance is a very strong clue, that there is
    no bond between those atoms, therefore the chain is discontinued.
    
    Parameters
    ----------
    array : AtomArray
        The array to be checked.
    min_len, max_len : float, optional
        The interval in which the atom-atom distance is evaluated as
        bond.
    
    Returns
    -------
    discontinuity : ndarray, dtype=bool
         Contains the indices of atoms after a discontinuity.
    """
    backbone_mask = filter_backbone(array)
    backbone = array[backbone_mask]
    diff = np.diff(backbone.coord, axis=0)
    sq_distance = np.sum(diff**2, axis=1)
    sq_min_len = min_len**2
    sq_max_len = max_len**2
    discon_mask = (sq_distance < sq_min_len) | (sq_distance > sq_max_len)
    # discon_mask is shortened by 1 due to np.diff
    # -> Lenthening by 1 to fit backbone
    # and shifting by one to get position after discontinuity
    discon_mask = np.concatenate(([False], discon_mask))
    # discon_mask is true for discontinuity in backbone
    # map it back to get discontinuity for original atom array
    discon_mask_full = np.zeros(array.array_length(), dtype=bool)
    discon_mask_full[backbone_mask] = discon_mask
    discontinuity = np.where(discon_mask_full)
    return discontinuity[0]


def check_duplicate_atoms(array):
    """
    Check if a structure contains duplicate atoms, i.e. two atoms in a
    structure have the same annotations (coordinates may be different).
    
    Duplicate atoms may appear, when a structure has occupancy for an
    atom at two or more positions or when the *altloc* positions are
    improperly read.
    
    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be checked.
    
    Returns
    -------
    duplicate : ndarray, dtype=bool
        Contains the indices of duplicate atoms.
        The first occurence of an atom is not counted as duplicate.
    """
    duplicates = []
    annots = [array.get_annotation(category) for category
              in array.get_annotation_categories()]
    for i in range(1, array.array_length()):
        # Start with assumption that all atoms in the array
        # until index i are duplicates of the atom at index i
        is_dublicate = np.full(i, True, dtype=bool)
        for annot in annots:
            # For each annotation array filter out the atoms until
            # index i that have an unequal annotation
            # to the atom at index i 
            is_dublicate &= (annot[:i] == annot[i])
        # After checking all annotation arrays,
        # if there still is any duplicate to the atom at index i,
        # add i the the list of duplicate atom indices
        if is_dublicate.any():
            duplicates.append(i)
    return np.array(duplicates)


def check_in_box(array):
    r"""
    Check if a structure contains atoms whose position is outside the
    box.

    Coordinates are outside the box, when they cannot be represented by
    a linear combination of the box vectors with scalar factors
    :math:`0 \le a_i \le 1`.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be checked.
    
    Returns
    -------
    outside : ndarray, dtype=bool
        Contains the indices of atoms outside the atom array's box.
    """
    if array.box is None:
        raise TypeError("Structure has no box")
    box = array.box
    fractions = coord_to_fraction(array, box)
    return np.where(((fractions >= 0) & (fractions < 1)).all(axis=-1))[0]


def renumber_atom_ids(array, start=None):
    """
    Renumber the atom IDs of the given array.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be checked.
    start : int, optional
        The starting index for renumbering.
        The first ID in the array is taken by default.

    Returns
    -------
    array : AtomArray or AtomArrayStack
        The renumbered array.
    """
    if "atom_id" not in array.get_annotation_categories():
        raise ValueError("The atom array must have the 'atom_id' annotation")
    if start is None:
        start = array.atom_id[0]
    array.atom_id = np.arange(start, array.shape[-1]+1)
    return array
    

def renumber_res_ids(array, start=None):
    """
    Renumber the residue IDs of the given array.

    Parameters
    ----------
    array : AtomArray or AtomArrayStack
        The array to be checked.
    start : int, optional
        The starting index for renumbering.
        The first ID in the array is taken by default.
    
    Returns
    -------
    array : AtomArray or AtomArrayStack
        The renumbered array.
    """
    if start is None:
        start = array.res_id[0]
    diff = np.diff(array.res_id)
    diff[diff != 0] = 1
    new_res_ids =  np.concatenate(([start], diff)).cumsum()
    array.res_id = new_res_ids
    return array



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.blast"
__author__ = "Patrick Kunzmann"
__all__ = ["BlastWebApp"]

from .alignment import BlastAlignment
from ..application import Application, requires_state, AppState
from ..webapp import WebApp, RuleViolationError
from ...sequence.sequence import Sequence
from ...sequence.seqtypes import NucleotideSequence, ProteinSequence
from ...sequence.io.fasta.file import FastaFile
from ...sequence.io.fasta.convert import get_sequence
from ...sequence.align.alignment import Alignment
import time
import requests
from xml.etree import ElementTree


_ncbi_url = "https://blast.ncbi.nlm.nih.gov/Blast.cgi"

class BlastWebApp(WebApp):
    """
    Perform a local alignment against a large sequence database using
    using the web-based BLAST application (by default NCBI BLAST).
    
    Parameters
    ----------
    program : str
        The specific BLAST program. One of 'blastn', 'megablast',
        'blastp', 'blastx', 'tblastn' and 'tblastx'.
    query : Sequence or str
        The query sequence. If a string is provided, it is interpreted
        as path to a FASTA file, if the string contains a valid FASTA
        file extension, otherwise it is interpreted as a single letter
        string representation of a sequence. 
    database : str, optional
        The NCBI sequence database to blast against. By default it
        contains all sequences (`database`='nr'`).
    app_url : str, optional
        URL of the BLAST web app. By default NCBI BLAST is used.
        This can be changed to a private server or another cloud
        provider.
    obey_rules : bool, optional
        If true, the application raises an :class:`RuleViolationError`,
        if the server is contacted too often, based on the NCBI BLAST
        usage rules. (Default: True)
    mail : str, optional
        If a mail address is provided, it will be appended in the
        HTTP request. This allows the NCBI to contact you in case
        your application sends too many requests.
    """
    
    _last_contact = 0
    _last_request = 0
    _contact_delay = 3
    _request_delay = 60
    
    def __init__(self, program, query, database="nr",
                 app_url=_ncbi_url, obey_rules=True,
                 mail="padix.key@gmail.com"):
        super().__init__(app_url, obey_rules)
        
        # 'megablast' is somehow not working
        # When entering the corresponding HTTPS request into a browser
        # you are redirected onto the blast mainpage
        if program not in ["blastn", "blastp",
                           "blastx", "tblastn", "tblastx"]:
            raise ValueError(f"'{program}' is not a valid BLAST program")
        self._program = program
        
        requires_protein = (program in ["blastp", "tblastn"])
        if isinstance(query, str) and query.endswith((".fa",".fst",".fasta")):
            # If string has a file extension, it is interpreted as
            # FASTA file from which the sequence is taken
            file = FastaFile.read(query)
            # Get first entry in file and take the sequence
            # (rather than header) 
            self._query = str(get_sequence(file))
        elif isinstance(query, Sequence):
            self._query = str(query)
        else:
            self._query = query
        
        # Check for unsuitable symbols in query string
        if requires_protein:
            ref_alphabet = ProteinSequence.alphabet
        else:
            ref_alphabet = NucleotideSequence.alphabet_amb
        for symbol in self._query:
            if not symbol.upper() in ref_alphabet:
                raise ValueError(
                    f"Query sequence contains unsuitable symbol {symbol}"
                )
        
        self._database = database
        
        self._gap_openining = None
        self._gap_extension = None
        self._word_size = None
        
        self._expect_value = None
        self._max_results = None
        self._entrez_query = None
        
        self._reward = None
        self._penalty = None
        
        self._matrix = None
        self._threshold = None
        
        self._mail=mail
        self._rid = None
    
    @requires_state(AppState.CREATED)
    def set_entrez_query(self, query):
        """
        Limit the size of the database.
        Only sequences that match the query are searched.

        Parameters
        ----------
        query : Query
            An NCBI Entrez query.
        """
        self._entrez_query = str(query)
    
    @requires_state(AppState.CREATED)
    def set_max_results(self, number):
        """
        Limit the maximum number of results.

        Parameters
        ----------
        number : int
            The maximum number of results.
        """
        self._max_results = number
    
    @requires_state(AppState.CREATED)
    def set_max_expect_value(self, value):
        """
        Set the threshold expectation value (E-value).
        No alignments with an E-value above this threshold will be
        considered.
        
        The E-Value is the expectation value for the number of random
        sequences of a similar sized database getting an equal or higher
        score by change when aligned with the query sequence.
        
        Parameters
        ----------
        value : float
            The threshold E-value.
        """
        self._expect_value = value
    
    @requires_state(AppState.CREATED)
    def set_gap_penalty(self, opening, extension):
        """
        Set the affine gap penalty for the alignment.
        
        Parameters
        ----------
        opening : float
            The penalty for gap opening.
        extension : float
            The penalty for gap extension.
        """
        self._gap_openining = opening
        self._gap_extension = extension
    
    @requires_state(AppState.CREATED)
    def set_word_size(self, size):
        """
        Set the word size for alignment seeds.
        
        Parameters
        ----------
        size : int
            Word size.
        """
        self._word_size = size
    
    @requires_state(AppState.CREATED)
    def set_match_reward(self, reward):
        """
        Set the score of a symbol match in the alignment.
        
        Used only in 'blastn' and 'megablast'.
        
        Parameters
        ----------
        reward : int
            Match reward. Must be positive.
        """
        self._reward = reward
    
    @requires_state(AppState.CREATED)
    def set_mismatch_penalty(self, penalty):
        """
        Set the penalty of a symbol mismatch in the alignment.
        
        Used only in 'blastn' and 'megablast'.
        
        Parameters
        ----------
        penalty : int
            Mismatch penalty. Must be negative.
        """
        self._penalty = penalty
    
    @requires_state(AppState.CREATED)
    def set_substitution_matrix(self, matrix_name):
        """
        Set the penalty of a symbol mismatch in the alignment.
        
        Used only in 'blastp', "blastx', 'tblastn' and 'tblastx'.
        
        Parameters
        ----------
        matrix_name : str
            Name of the substitution matrix. Default is 'BLOSUM62'.
        """
        self._matrix = matrix_name.upper()
    
    @requires_state(AppState.CREATED)
    def set_threshold(self, threshold):
        """
        Set the threshold neighboring score for initial words.
        
        Used only in 'blastp', "blastx', 'tblastn' and 'tblastx'.
        
        Parameters
        ----------
        threshold : int
            Threshold value. Must be positve.
        """
        self._threshold = threshold
    
    def run(self):
        param_dict = {}
        param_dict["tool"] = "Biotite"
        param_dict["email"] = self._mail
        param_dict["CMD"] = "Put"
        param_dict["PROGRAM"] = self._program
        param_dict["QUERY"] = str(self._query)
        param_dict["DATABASE"] = self._database
        if self._entrez_query is not None:
            param_dict["ENTREZ_QUERY"] = self._entrez_query
        if self._max_results is not None:
            param_dict["HITLIST_SIZE"] = str(self._max_results)
        if self._expect_value is not None:
            param_dict["EXPECT"] = self._expect_value
        if self._gap_openining is not None and self._gap_extension is not None:
            param_dict["GAPCOSTS"] = "{:d} {:d}".format(self._gap_openining,
                                                      self._gap_extension)
        if self._word_size is not None:
            param_dict["WORD_SIZE"] = self._word_size
        
        if self._program in ["blastn", "megablast"]:
            if self._reward is not None:
                param_dict["NUCL_REWARD"] = self._reward
            if self._penalty is not None:
                param_dict["NUCL_PENALTY"] = self._penalty
        
        if self._program in ["blastp", "blastx", "tblastn", "tblastx"]:
            if self._matrix is not None:
                param_dict["MATRIX"] = self._matrix
            if self._threshold is not None:
                param_dict["THRESHOLD"] = self._threshold
        
        request = requests.get(self.app_url(), params=param_dict)
        if "Submitted URI too large" in request.text:
            raise ValueError("The URI is too large, try a shorter sequence")
        self._contact()
        self._request()
        info_dict = BlastWebApp._get_info(request.text)
        self._rid = info_dict["RID"]
    
    def is_finished(self):
        data_dict = {"FORMAT_OBJECT" : "SearchInfo",
                     "RID"           : self._rid,
                     "CMD"           : "Get"}
        request = requests.get(self.app_url(), params=data_dict)
        self._contact()
        info_dict = BlastWebApp._get_info(request.text)
        if info_dict["Status"] == "UNKNOWN":
            # Indicates invalid query input values
            raise ValueError(
                "The input values seem to be invalid "
                "(Server responsed status 'UNKNOWN')"
            )
        return info_dict["Status"] == "READY"
        
    def wait_interval(self):
        # NCBI requires a 3 second delay between server contacts
        return BlastWebApp._contact_delay
    
    def clean_up(self):
        param_dict = {}
        param_dict["CMD"] = "Delete"
        param_dict["RID"] = self._rid
        request = requests.get(self.app_url(), params=param_dict)
    
    def evaluate(self):
        param_dict = {}
        param_dict["tool"] = "BiotiteClient"
        if self._mail is not None:
            param_dict["email"] = self._mail
        param_dict["CMD"] = "Get"
        param_dict["RID"] = self._rid
        param_dict["FORMAT_TYPE"] = "XML"
        param_dict["NCBI_GI"] = "T"
        request = requests.get(self.app_url(), params=param_dict)
        self._contact()
        
        self._alignments = []
        self._xml_response = request.text
        root = ElementTree.fromstring(self._xml_response)
        # Extract BlastAlignment objects from <Hit> tags
        hit_xpath = "./BlastOutput_iterations/Iteration/Iteration_hits/Hit"
        hits = root.findall(hit_xpath)
        for hit in hits:
            hit_definition = hit.find("Hit_def").text
            hit_id = hit.find("Hit_accession").text
            hsp = hit.find(".Hit_hsps/Hsp")
            score = int(hsp.find("Hsp_score").text)
            e_value = float(hsp.find("Hsp_evalue").text)
            query_begin = int(hsp.find("Hsp_query-from").text)
            query_end = int(hsp.find("Hsp_query-to").text)
            hit_begin = int(hsp.find("Hsp_hit-from").text)
            hit_end = int(hsp.find("Hsp_hit-to").text)
            
            seq1_str = hsp.find("Hsp_qseq").text
            seq2_str = hsp.find("Hsp_hseq").text
            if self._program in ["blastn", "megablast"]:
                # NucleotideSequence/ProteinSequence do ignore gaps
                # Gaps are represented by the trace
                seq1, seq2 = [
                    NucleotideSequence(s.replace("-", ""))
                    for s in (seq1_str, seq2_str)
                ]
            else:
                seq1, seq2 = [
                    ProteinSequence(s.replace("-", "").replace("U", "C"))
                    for s in (seq1_str, seq2_str)
                ]
            trace = Alignment.trace_from_strings([seq1_str, seq2_str])
            
            alignment = BlastAlignment( [seq1 ,seq2], trace, score, e_value,
                                        (query_begin, query_end),
                                        (hit_begin, hit_end),
                                        hit_id, hit_definition )
            self._alignments.append(alignment)

    @requires_state(AppState.JOINED)
    def get_xml_response(self):
        """
        Get the raw XML response.
        
        Returns
        -------
        response : str
            The raw XML response.
        """
        return self._xml_response

    @requires_state(AppState.JOINED)
    def get_alignments(self):
        """
        Get the resulting local sequence alignments.
        
        Returns
        -------
        alignment : list of BlastAlignment
            The local sequence alignments.
        """
        return self._alignments
    
    @staticmethod
    def _get_info(text):
        """
        Get the *QBlastInfo* block of the response HTML as dictionary
        """
        lines = [line for line in text.split("\n")]
        info_dict = {}
        in_info_block = False
        for line in lines:
            if "QBlastInfoBegin" in line:
                in_info_block = True
                continue
            if "QBlastInfoEnd" in line:
                in_info_block = False
                continue
            if in_info_block:
                pair = line.split("=")
                info_dict[pair[0].strip()] = pair[1].strip()
        return info_dict
    
    def _contact(self):
        """
        Resets the time since the last server contact. Used for
        detecting server rule violation.
        """
        contact = time.time()
        if (contact - BlastWebApp._last_contact) < BlastWebApp._contact_delay:
            self.violate_rule("The server was contacted too often")
        BlastWebApp._last_contact = contact
    
    def _request(self):
        """
        Resets the time since the last new alignment request. Used for
        detecting server rule violation.
        """
        request = time.time()
        if (request - BlastWebApp._last_request) < BlastWebApp._request_delay:
            self.violate_rule("Too frequent BLAST requests")
        BlastWebApp._last_request = request



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import pytest
import numpy as np
import biotite.structure as struc
import biotite.structure.io.pdb as pdb
import biotite.structure.io.mmtf as mmtf
from ..util import data_dir, cannot_import


# Ignore warning about dummy unit cell vector
@pytest.mark.filterwarnings("ignore")
@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize("pdb_id", ["1l2y", "1gya"])
def test_single(pdb_id):
    file_name = join(data_dir("structure"), pdb_id+".pdb")
    
    # Single atom SASA, compare with MDTraj
    file = pdb.PDBFile.read(file_name)
    array = file.get_structure(model=1)
    sasa = struc.sasa(array, vdw_radii="Single", point_number=5000)
    
    from biotite.structure.info.radii import _SINGLE_RADII as radii
    import mdtraj
    # Use the same atom radii
    radii = {element.capitalize() : radius / 10
             for element, radius in radii.items()}
    traj = mdtraj.load(file_name)
    # Conversion from nm^2 to A^2
    sasa_exp = mdtraj.shrake_rupley(
        traj, change_radii=radii, n_sphere_points=5000
    )[0] * 100

    
    # Assert that more than 90% of atoms
    # have less than 10% SASA difference
    assert np.count_nonzero(
        np.isclose(sasa, sasa_exp, rtol=1e-1, atol=1e-1)
    ) / len(sasa) > 0.9
    # Assert that more than 98% of atoms
    # have less than 1% SASA difference
    assert np.count_nonzero(
        np.isclose(sasa, sasa_exp, rtol=1e-2, atol=1e-1)
    ) / len(sasa) > 0.98


@pytest.mark.parametrize("pdb_id", ["1l2y", "1gya"])
def test_coarse_grained(pdb_id):
    # Multi atom SASA (ProtOr), compare with single atom SASA
    # on residue level
    file = mmtf.MMTFFile.read(join(data_dir("structure"), pdb_id+".mmtf"))
    array = mmtf.get_structure(file, model=1)
    array = array[struc.filter_amino_acids(array)]
    sasa = struc.apply_residue_wise(
        array, struc.sasa(array, vdw_radii="ProtOr"), np.nansum
    )
    sasa_exp = struc.apply_residue_wise(
        array, struc.sasa(array, vdw_radii="Single"), np.nansum
    )

    # Assert that more than 90% of atoms
    # have less than 10% SASA difference
    assert np.count_nonzero(
        np.isclose(sasa, sasa_exp, rtol=1e-1, atol=1)
    ) / len(sasa) > 0.9
    # Assert that more than 98% of atoms
    # have less than 40% SASA difference
    assert np.count_nonzero(
        np.isclose(sasa, sasa_exp, rtol=4e-1, atol=1)
    ) / len(sasa) > 0.98


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pytest
import json
import numpy as np
import pickle as pkl
import biotite.structure as struc
import biotite.structure.io as strucio
from os.path import join
from ..util import data_dir


@pytest.fixture
def nuc_sample_array():
    """
    Sample structure for pseudoknot detection.
    """
    return strucio.load_structure(join(data_dir("structure"), "4p5j.cif"))

def test_pseudoknots(nuc_sample_array):
    """
    Check the output of :func:`pseudoknots()`.
    """
    # Known base pairs with pseudoknot-order = 1:
    pseudoknot_order_one = [{2, 74}, {58, 72}, {59, 71}, {60, 70}]
    # Known base pairs that can either be of order one or two
    pseudoknot_order_one_or_two =  [{9, 48}, {10, 49}]
    order_one_count = (
        len(pseudoknot_order_one) + (len(pseudoknot_order_one_or_two)/2)
    )
    order_two_count = len(pseudoknot_order_one_or_two)/2

    base_pairs = struc.base_pairs(nuc_sample_array)
    pseudoknot_order = struc.pseudoknots(base_pairs)

    # Sample structure should have two optimal solutions with default
    # scoring parameters
    assert len(pseudoknot_order) == 2

    for optimal_solution in pseudoknot_order:
        # Assert that the right number of pseudoknots is present for
        # each order
        assert len(base_pairs) == len(optimal_solution)
        assert np.count_nonzero(optimal_solution == 1) == order_one_count
        assert np.count_nonzero(optimal_solution == 2) == order_two_count
        assert np.max(optimal_solution) == 2

        # Assert that the each base pair has the right pseudoknot order
        for base_pair, order in zip(
            nuc_sample_array[base_pairs].res_id, optimal_solution
        ):
            if(order == 1):
                assert (
                    set(base_pair) in pseudoknot_order_one or
                    set(base_pair) in pseudoknot_order_one_or_two
                )
            elif (order == 2):
                assert (
                    set(base_pair) in pseudoknot_order_one_or_two
                )

def load_test(name):
    """
    Load sample base pair arrays and reference solutions from file.
    """
    # Base pairs as numpy array (input for `pseudoknots()`)
    with open(
        join(data_dir("structure"), "pseudoknots", f"{name}_knotted.json"),
        "r"
    ) as f:
        basepairs = np.array(json.load(f))
    # List of solutions (set of tuples)
    with open(
        join(data_dir("structure"), "pseudoknots", f"{name}_unknotted.json"),
        "rb"
    ) as f:
        solutions = json.load(f)
    for i, solution in enumerate(solutions):
        solutions[i] = set([tuple(pair) for pair in solution])
    return basepairs, solutions

@pytest.mark.parametrize("name", [f"test{x}" for x in range(21)])
def test_pseudoknot_removal(name):
    """
    Test the implementation of the dynamic programming algorithm
    referenced in :func:`pseudoknots()` against the original
    implementation.

    The reference solutions were created with the following tool:
    https://www.ibi.vu.nl/programs/k2nwww/

    The original purpose was to remove pseudoknots. Thus, the reference
    solutions contain unknotted basepairs.
    """
    # Get base pairs and reference solutions
    basepairs, reference_solutions = load_test(name)

    # Calculate solutions from the base pairs
    raw_solutions = struc.pseudoknots(basepairs, max_pseudoknot_order=0)

    # The number of solutions calculated
    solutions_count = 0

    # Verify that each solution is in the reference solutions
    for raw_solution in raw_solutions:
        solution = set()
        for basepair, order in zip(basepairs, raw_solution):
            if order == -1:
                continue
            solution.add(tuple(sorted(basepair)))
        solutions_count += 1
        assert solution in reference_solutions

    # Verify that the number of solutions matches the reference
    assert len(reference_solutions) == solutions_count

@pytest.mark.parametrize("seed", range(10))
def test_pseudoknot_orders(seed):
    """
    Generate a random set of basepairs. Assert that increasing
    pseudoknot orders contain less or equal base pairs. Furthermore,
    assert that each individual order only contains unknotted base
    pairs.
    """
    # Generate Random set of basepairs
    np.random.seed(seed)
    bases = range(100)
    basepairs = np.random.choice(bases, size=(20, 2), replace=False)

    # Get pseudoknot order for each basepair
    solutions = struc.pseudoknots(basepairs)

    # Iterate through the solutions
    for solution in solutions:
        # Number of base pairs in the previous order
        previous_order = -1
        for order in range(np.max(solution)+1):
            # Ensure that the base pairs of the same order are unknotted
            assert (struc.pseudoknots(basepairs[solution == order]) == 0).all()

            # Number of base pairs in the current order
            this_order = len(solution[solution == order])
            # Ensure that that higher orders contain less or equal base
            # pairs than lower orders
            if previous_order != -1:
                assert this_order <= previous_order
            previous_order = this_order



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.netcdf"
__author__ = "Patrick Kunzmann"
__all__ = ["NetCDFFile"]

import numpy as np
from ..trajfile import TrajectoryFile
from ...box import vectors_from_unitcell, unitcell_from_vectors


class NetCDFFile(TrajectoryFile):
    """
    This file class represents a NetCDF trajectory file.
    """
    
    @classmethod
    def traj_type(cls):
        import mdtraj.formats as traj
        return traj.NetCDFTrajectoryFile
    
    @classmethod
    def process_read_values(cls, read_values):
        # .dcd files use Angstrom
        coord = read_values[0]
        time = read_values[1]
        cell_lengths = read_values[2]
        cell_angles = read_values[3]
        if cell_lengths is None or cell_angles is None:
             box = None
        else:
            box = np.stack(
                [vectors_from_unitcell(a, b, c, alpha, beta, gamma)
                for (a, b, c), (alpha, beta, gamma)
                in zip(cell_lengths, np.deg2rad(cell_angles))],
                axis=0
            )
        return coord, box, time
    
    @classmethod
    def prepare_write_values(cls, coord, box, time):
        coord = coord.astype(np.float32, copy=False) \
              if coord is not None else None
        time = time.astype(np.float32, copy=False) \
               if time is not None else None
        if box is None:
            cell_lengths = None
            cell_angles  = None
        else:
            cell_lengths = np.zeros((len(box), 3), dtype=np.float32)
            cell_angles  = np.zeros((len(box), 3), dtype=np.float32)
            for i, model_box in enumerate(box):
                a, b, c, alpha, beta, gamma = unitcell_from_vectors(model_box)
                cell_lengths[i] = np.array((a, b, c))
                cell_angles[i] = np.rad2deg((alpha, beta, gamma))
        return {
            "coordinates" : coord,
            "time" : time,
            "cell_lengths" : cell_lengths,
            "cell_angles" : cell_angles,
        }


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides functions for calculation of characteristic values when
comparing multiple structures with each other.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"
__all__ = ["rmsd", "rmsf", "average"]

import numpy as np
from .atoms import Atom, AtomArray, AtomArrayStack, coord
from .util import vector_dot


def rmsd(reference, subject):
    r"""
    Calculate the RMSD between two structures.
    
    The *root-mean-square-deviation* (RMSD) indicates the overall
    deviation of each model of a structure to a reference structure.
    It is defined as:
    
    .. math:: RMSD = \sqrt{ \frac{1}{n} \sum\limits_{i=1}^n (x_i - x_{ref,i})^2}
    
    Parameters
    ----------
    reference : AtomArray or ndarray, dtype=float, shape=(n,3)
        The reference structure.
        Alternatively, coordinates can be provided directly as
        :class:`ndarray`.
    subject : AtomArray or AtomArrayStack or ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        Structure(s) to be compared with `reference`.
        Alternatively, coordinates can be provided directly as
        :class:`ndarray`.
    
    Returns
    -------
    rmsd : float or ndarray, dtype=float, shape=(m,)
        RMSD between subject and reference.
        If subject is an :class:`AtomArray` a float is returned.
        If subject is an :class:`AtomArrayStack` a :class:`ndarray`
        containing the RMSD for each model is returned.
    
    See Also
    --------
    rmsf

    Notes
    -----
    This function does not superimpose the subject to its reference.
    In most cases :func:`superimpose()` should be called prior to this
    function.

    Examples
    --------

    Calculate the RMSD of all models to the first model:

    >>> superimposed, _ = superimpose(atom_array, atom_array_stack)
    >>> rms = rmsd(atom_array, superimposed)
    >>> print(np.around(rms, decimals=3))
    [0.000 1.928 2.103 2.209 1.806 2.172 2.704 1.360 2.337 1.818 1.879 2.471
     1.939 2.035 2.167 1.789 1.653 2.348 2.247 2.529 1.583 2.115 2.131 2.050
     2.512 2.666 2.206 2.397 2.328 1.868 2.316 1.984 2.124 1.761 2.642 1.721
     2.571 2.579]
    """
    return np.sqrt(np.mean(_sq_euclidian(reference, subject), axis=-1))


def rmsf(reference, subject):
    r"""
    Calculate the RMSF between two structures.

    The *root-mean-square-fluctuation* (RMSF) indicates the positional
    deviation of a structure to a reference structure, averaged over all
    models.
    Usually the reference structure, is the average over all models.
    The RMSF is defined as:
    
    .. math:: RMSF(i) = \sqrt{ \frac{1}{T} \sum\limits_{t=1}^T (x_i(t) - x_{ref,i}(t))^2}
    
    Parameters
    ----------
    reference : AtomArray or ndarray, dtype=float, shape=(n,3)
        The reference structure.
        Alternatively, coordinates can be provided directly as
        :class:`ndarray`.
    subject : AtomArrayStack or ndarray, dtype=float, shape=(m,n,3)
        Structures to be compared with `reference`.
        The time *t* is represented by the models in the
        :class:`AtomArrayStack`.
        Alternatively, coordinates can be provided directly as
        :class:`ndarray`.
    
    Returns
    -------
    rmsf : ndarray, dtype=float, shape=(n,)
        RMSF between subject and reference structure.
        Each element gives the RMSF for the atom at the respective
        index.
    
    See Also
    --------
    rmsd

    Notes
    -----
    This function does not superimpose the subject to its reference.
    In most cases :func:`superimpose()` should be called prior to this
    function.

    Examples
    --------

    Calculate the :math:`C_\alpha` RMSF of all models to the average
    model:

    >>> ca = atom_array_stack[:, atom_array_stack.atom_name == "CA"]
    >>> ca_average = average(ca)
    >>> ca, _ = superimpose(ca_average, ca)
    >>> print(rmsf(ca_average, ca))
    [1.372 0.360 0.265 0.261 0.288 0.204 0.196 0.306 0.353 0.238 0.266 0.317
     0.358 0.448 0.586 0.369 0.332 0.396 0.410 0.968]
    """
    return np.sqrt(np.mean(_sq_euclidian(reference, subject), axis=-2))


def average(atoms):
    """
    Calculate an average structure.
    
    The average structure has the average coordinates
    of the input models.
    
    Parameters
    ----------
    atoms : AtomArrayStack or ndarray, dtype=float, shape=(m,n,3)
        The structure models to be averaged.
        Alternatively, coordinates can be provided directly as
        :class:`ndarray`.
    
    Returns
    -------
    average : AtomArray or ndarray, dtype=float, shape=(n,3)
        Structure with averaged atom coordinates.
        If `atoms` is a :class:`ndarray` and :class:`ndarray` is also
        returned.
    
    See Also
    --------
    rmsd, rmsf
    
    Notes
    -----
    The calculated average structure is not suitable for visualization
    or geometric calculations, since bond lengths and angles will
    deviate from meaningful values.
    This method is rather useful to provide a reference structure for
    calculation of e.g. the RMSD or RMSF. 
    """
    coords = coord(atoms)
    if coords.ndim != 3:
        raise TypeError(
            "Expected an AtomArrayStack or an ndarray with shape (m,n,3)"
        )
    mean_coords = np.mean(coords, axis=0)
    if isinstance(atoms, AtomArrayStack):
        mean_array = atoms[0].copy()
        mean_array.coord = mean_coords
        return mean_array
    else:
        return mean_coords


def _sq_euclidian(reference, subject):
    """
    Calculate squared euclidian distance between atoms in two
    structures.
    
    Parameters
    ----------
    reference : AtomArray or ndarray, dtype=float, shape=(n,3)
        Reference structure.
    subject : AtomArray or AtomArrayStack or ndarray, dtype=float, shape=(n,3) or shape=(m,n,3)
        Structure(s) whose atoms squared euclidian distance to
        `reference` is measured.
    
    Returns
    -------
    ndarray, dtype=float, shape=(n,) or shape=(m,n)
        Squared euclidian distance between subject and reference.
        If subject is an :class:`AtomArray` a 1-D array is returned.
        If subject is an :class:`AtomArrayStack` a 2-D array is
        returned.
        In this case the first dimension indexes the AtomArray.
    """
    reference_coord = coord(reference)
    subject_coord = coord(subject)
    if reference_coord.ndim != 2:
        raise TypeError(
            "Expected an AtomArray or an ndarray with shape (n,3) as reference"
        )
    dif = subject_coord - reference_coord
    return vector_dot(dif, dif)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.align"
__author__ = "Patrick Kunzmann"

from ..sequence import Sequence
from ..seqtypes import NucleotideSequence, ProteinSequence
from ..alphabet import Alphabet
import numpy as np
import os

__all__ = ["SubstitutionMatrix"]


class SubstitutionMatrix(object):
    """
    A :class:`SubstitutionMatrix` is the foundation for scoring in
    sequence alignments.
    A :class:`SubstitutionMatrix` maps each possible pairing of a symbol
    of a first alphabet with a symbol of a second alphabet to a score
    (integer).
    
    The class uses a 2-D (m x n) :class:`ndarray`
    (dtype=:attr:`numpy.int32`),
    where each element stores the score for a symbol pairing, indexed
    by the symbol codes of the respective symbols in an *m*-length
    alphabet 1 and an *n*-length alphabet 2.
    
    There are 3 ways to creates instances:
    
    At first a 2-D :class:`ndarray` containing the scores can be
    directly provided.
    
    Secondly a dictionary can be provided, where the keys are pairing
    tuples and values are the corresponding scores.
    The pairing tuples consist of a symbol of alphabet 1 as first
    element and a symbol of alphabet 2 as second element. Parings have
    to be provided for each possible combination.
    
    At last a valid matrix name can be given, which is loaded from the
    internal matrix database. The following matrices are avaliable:
        
        - Nucleotide substitution matrices from NCBI database
            - **NUC** - Also usable with ambiguous alphabet
        
        - Protein substitution matrices from NCBI database
        
            - **PAM<n>**
            - **BLOSUM<n>**
            - **MATCH** - Only differentiates between match and mismatch
            - **IDENTITY** - Strongly penalizes mismatches
            - **GONNET** - Not usable with default protein alphabet
            - **DAYHOFF**
        
        - Corrected protein substitution matrices :footcite:`Hess2016`,
          **<BLOCKS>** is the BLOCKS version, the matrix is based on
            
            - **BLOSUM<n>_<BLOCKS>**
            - **RBLOSUM<n>_<BLOCKS>**
            - **CorBLOSUM<n>_<BLOCKS>**
    
    A list of all available matrix names is returned by
    :meth:`list_db()`.
    
    Since this class can handle two different alphabets, it is possible
    to align two different types of sequences.
    
    Objects of this class are immutable.
    
    Parameters
    ----------
    alphabet1 : Alphabet, length=m
        The first alphabet of the substitution matrix.
    alphabet2 : Alphabet, length=n
        The second alphabet of the substitution matrix.
    score_matrix : ndarray, shape=(m,n) or dict or str
        Either a symbol code indexed :class:`ndarray` containing the scores,
        or a dictionary mapping the symbol pairing to scores,
        or a string referencing a matrix in the internal database.
    
    Raises
    ------
    KeyError
        If the matrix dictionary misses a symbol given in the alphabet.
    
    References
    ----------
    
    .. footbibliography::
    
    Examples
    --------
    
    Creating a matrix for two different (nonsense) alphabets
    via a matrix dictionary:
    
    >>> alph1 = Alphabet(["foo","bar"])
    >>> alph2 = Alphabet([1,2,3])
    >>> matrix_dict = {("foo",1):5,  ("foo",2):10, ("foo",3):15,
    ...                ("bar",1):42, ("bar",2):42, ("bar",3):42}
    >>> matrix = SubstitutionMatrix(alph1, alph2, matrix_dict)
    >>> print(matrix.score_matrix())
    [[ 5 10 15]
     [42 42 42]]
    >>> print(matrix.get_score("foo", 2))
    10
    >>> print(matrix.get_score_by_code(0, 1))
    10

    Creating an identity substitution matrix via the score matrix:

    >>> alph = NucleotideSequence.alphabet_unamb
    >>> matrix = SubstitutionMatrix(alph, alph, np.identity(len(alph)))
    >>> print(matrix)
        A   C   G   T
    A   1   0   0   0
    C   0   1   0   0
    G   0   0   1   0
    T   0   0   0   1
    
    Creating a matrix via database name:
        
    >>> alph = ProteinSequence.alphabet
    >>> matrix = SubstitutionMatrix(alph, alph, "BLOSUM50")
    """
    
    # Directory of matrix files
    _db_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                           "matrix_data")
    
    def __init__(self, alphabet1, alphabet2, score_matrix):
        self._alph1 = alphabet1
        self._alph2 = alphabet2
        if isinstance(score_matrix, dict):
            self._fill_with_matrix_dict(score_matrix)
        elif isinstance(score_matrix, np.ndarray):
            alph_shape = (len(alphabet1), len(alphabet2))
            if score_matrix.shape != alph_shape:
                raise ValueError(
                    f"Matrix has shape {score_matrix.shape}, "
                    f"but {alph_shape} is required"
                )
            self._matrix = score_matrix.astype(np.int32)
        elif isinstance(score_matrix, str):
            matrix_dict = SubstitutionMatrix.dict_from_db(score_matrix)
            self._fill_with_matrix_dict(matrix_dict)
        else:
            raise TypeError("Matrix must be either a dictionary, "
                            "an 2-D ndarray or a string")
        # This class is immutable and has a getter function for the
        # score matrix -> make the score matrix read-only
        self._matrix.setflags(write=False)

    def __repr__(self):
        """Represent SubstitutionMatrix as a string for debugging."""
        return f"SubstitutionMatrix({self._alph1.__repr__()}, {self._alph2.__repr__()}, " \
               f"np.{np.array_repr(self._matrix)})"

    def __eq__(self, item):
        if not isinstance(item, SubstitutionMatrix):
            return False
        if self._alph1 != item.get_alphabet1():
            return False
        if self._alph2 != item.get_alphabet2():
            return False
        if not np.array_equal(self.score_matrix(), item.score_matrix()):
            return False
        return True

    def __ne__(self, item):
        return not self == item

    def _fill_with_matrix_dict(self, matrix_dict):
        self._matrix = np.zeros(( len(self._alph1), len(self._alph2) ),
                                dtype=np.int32)
        for i in range(len(self._alph1)):
            for j in range(len(self._alph2)):
                sym1 = self._alph1.decode(i)
                sym2 = self._alph2.decode(j)
                self._matrix[i,j] = int(matrix_dict[sym1, sym2])
    
    def get_alphabet1(self):
        """
        Get the first alphabet. 
        
        Returns
        -------
        alphabet : Alphabet
            The first alphabet.
        """
        return self._alph1
    
    def get_alphabet2(self):
        """
        Get the second alphabet. 
        
        Returns
        -------
        alphabet : Alphabet
            The second alphabet.
        """
        return self._alph2
    
    def score_matrix(self):
        """
        Get the 2-D :class:`ndarray` containing the score values.
        
        Returns
        -------
        matrix : ndarray, shape=(m,n), dtype=np.int32
            The symbol code indexed score matrix.
            The array is read-only.
        """
        return self._matrix
    
    def transpose(self):
        """
        Get a copy of this instance, where the alphabets are
        interchanged.
        
        Returns
        -------
        transposed : SubstitutionMatrix
            The transposed substitution matrix.
        """
        new_alph1 = self._alph2
        new_alph2 = self._alph1
        new_matrix = np.transpose(self._matrix)
        return SubstitutionMatrix(new_alph1, new_alph2, new_matrix)
    
    def is_symmetric(self):
        """
        Check whether the substitution matrix is symmetric,
        i.e. both alphabets are identical
        and the score matrix is symmetric.

        Returns
        -------
        is_symmetric : bool
            True, if both alphabets are identical and the score matrix
            is symmetric, false otherwise.
        """
        return     self._alph1 == self._alph2 \
               and np.array_equal(self._matrix, np.transpose(self._matrix))
    
    def get_score_by_code(self, code1, code2):
        """
        Get the substitution score of two symbols,
        represented by their code.
        
        Parameters
        ----------
        code1, code2 : int
            Symbol codes of the two symbols to be aligned.
        
        Returns
        -------
        score : int
            The substitution / alignment score.
        """
        return self._matrix[code1, code2]
    
    def get_score(self, symbol1, symbol2):
        """
        Get the substitution score of two symbols.
        
        Parameters
        ----------
        symbol1, symbol2 : object
            Symbols to be aligned.
        
        Returns
        -------
        score : int
            The substitution / alignment score.
        """
        code1 = self._alph1.encode(symbol1)
        code2 = self._alph2.encode(symbol2)
        return self._matrix[code1, code2]
    
    def shape(self):
        """
        Get the shape (i.e. the length of both alphabets)
        of the subsitution matrix.
        
        Returns
        -------
        shape : tuple
            Matrix shape.
        """
        return (len(self._alph1), len(self._alph2))
    
    def __str__(self):
        # Create matrix in NCBI format
        string = " "
        for symbol in self._alph2:
            string += f" {symbol:>3}"
        string += "\n"
        for i, symbol in enumerate(self._alph1):
            string += f"{symbol:>1}"
            for j in range(len(self._alph2)):
                string += f" {int(self._matrix[i,j]):>3d}"
            string += "\n"
        # Remove terminal line break
        string = string[:-1]
        return string
    
    @staticmethod
    def dict_from_str(string):
        """
        Create a matrix dictionary from a string in NCBI matrix format.
        
        Symbols of the first alphabet are taken from the left column,
        symbols of the second alphabet are taken from the top row.
        
        The keys of the dictionary consist of tuples containing the
        aligned symbols and the values are the corresponding scores.
        
        Returns
        -------
        matrix_dict : dict
            A dictionary representing the substitution matrix.
        """
        lines = [line.strip() for line in string.split("\n")]
        lines = [line for line in lines if len(line) != 0 and line[0] != "#"]
        symbols1 = [line.split()[0] for line in lines[1:]]
        symbols2 = [e for e in lines[0].split()]
        scores = np.array([line.split()[1:] for line in lines[1:]]).astype(int)
        scores = np.transpose(scores)
        
        matrix_dict = {}
        for i in range(len(symbols1)):
            for j in range(len(symbols2)):
                matrix_dict[(symbols1[i], symbols1[j])] = scores[i,j]
        return matrix_dict
    
    @staticmethod
    def dict_from_db(matrix_name):
        """
        Create a matrix dictionary from a valid matrix name in the
        internal matrix database.
        
        The keys of the dictionary consist of tuples containing the
        aligned symbols and the values are the corresponding scores.
        
        Returns
        -------
        matrix_dict : dict
            A dictionary representing the substitution matrix.
        """
        filename = SubstitutionMatrix._db_dir + os.sep + matrix_name + ".mat"
        with open(filename, "r") as f:
            return SubstitutionMatrix.dict_from_str(f.read())
    
    @staticmethod
    def list_db():
        """
        List all matrix names in the internal database.
        
        Returns
        -------
        db_list : list
            List of matrix names in the internal database.
        """
        files = os.listdir(SubstitutionMatrix._db_dir)
        # Remove '.mat' from files
        return [file[:-4] for file in sorted(files)]
        
    
    @staticmethod
    def std_protein_matrix():
        """
        Get the default :class:`SubstitutionMatrix` for protein sequence
        alignments, which is BLOSUM62.
        
        Returns
        -------
        matrix : SubstitutionMatrix
            Default matrix.
        """
        return _matrix_blosum62
    
    @staticmethod
    def std_nucleotide_matrix():
        """
        Get the default :class:`SubstitutionMatrix` for DNA sequence
        alignments.
        
        Returns
        -------
        matrix : SubstitutionMatrix
            Default matrix.
        """
        return _matrix_nuc

# Preformatted BLOSUM62 and NUC substitution matrix from NCBI
_matrix_blosum62 = SubstitutionMatrix(ProteinSequence.alphabet,
                                      ProteinSequence.alphabet,
                                      "BLOSUM62")
_matrix_nuc = SubstitutionMatrix(NucleotideSequence.alphabet_amb,
                                 NucleotideSequence.alphabet_amb,
                                 "NUC")




# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.database.rcsb"
__author__ = "Patrick Kunzmann"
__all__ = ["fetch"]

import requests
from os.path import isdir, isfile, join, getsize
import os
import glob
import io
from ..error import RequestError


_standard_url = "https://files.rcsb.org/download/"
_mmtf_url = "https://mmtf.rcsb.org/v1.0/full/"
_fasta_url = "https://www.rcsb.org/fasta/entry/"

_binary_formats = ["mmtf"]


def fetch(pdb_ids, format, target_path=None, overwrite=False, verbose=False):
    """
    Download structure files (or sequence files) from the RCSB PDB in
    various formats.
    
    This function requires an internet connection.
    
    Parameters
    ----------
    pdb_ids : str or iterable object of str
        A single PDB ID or a list of PDB IDs of the structure(s)
        to be downloaded .
    format : {'pdb', 'pdbx', 'cif', 'mmcif', 'mmtf', 'fasta'}
        The format of the files to be downloaded.
        ``'pdbx'``, ``'cif'`` and ``'mmcif'`` are synonyms for
        the same format.
    target_path : str, optional
        The target directory of the downloaded files.
        By default, the file content is stored in a file-like object
        (`StringIO` or `BytesIO`, respectively).
    overwrite : bool, optional
        If true, existing files will be overwritten. Otherwise the
        respective file will only be downloaded if the file does not
        exist yet in the specified target directory or if the file is
        empty. (Default: False)
    verbose: bool, optional
        If true, the function will output the download progress.
        (Default: False)
    
    Returns
    -------
    files : str or StringIO or BytesIO or list of (str or StringIO or BytesIO)
        The file path(s) to the downloaded files.
        If a single string (a single ID) was given in `pdb_ids`,
        a single string is returned. If a list (or other iterable
        object) was given, a list of strings is returned.
        If no `target_path` was given, the file contents are stored in
        either `StringIO` or `BytesIO` objects.
    
    Warnings
    --------
    Even if you give valid input to this function, in rare cases the
    database might return no or malformed data to you.
    In these cases the request should be retried.
    When the issue occurs repeatedly, the error is probably in your
    input.
    
    Examples
    --------
    
    >>> import os.path
    >>> file = fetch("1l2y", "cif", path_to_directory)
    >>> print(os.path.basename(file))
    1l2y.cif
    >>> files = fetch(["1l2y", "3o5r"], "cif", path_to_directory)
    >>> print([os.path.basename(file) for file in files])
    ['1l2y.cif', '3o5r.cif']
    """
    # If only a single PDB ID is present,
    # put it into a single element list
    if isinstance(pdb_ids, str):
        pdb_ids = [pdb_ids]
        single_element = True
    else:
        single_element = False
    # Create the target folder, if not existing
    if target_path is not None and not os.path.isdir(target_path):
        os.makedirs(target_path)
    
    files = []
    for i, id in enumerate(pdb_ids):
        # Verbose output
        if verbose:
            print(f"Fetching file {i+1:d} / {len(pdb_ids):d} ({id})...",
                  end="\r")
        
        # Fetch file from database
        if target_path is not None:
            file = join(target_path, id + "." + format)
        else:
            # 'file = None' -> store content in a file-like object
            file = None
        
        if file is None \
           or not isfile(file) \
           or getsize(file) == 0 \
           or overwrite:
                if format == "pdb":
                    r = requests.get(_standard_url + id + ".pdb")
                    content = r.text
                    _assert_valid_file(content, id)
                elif format in ["cif", "mmcif", "pdbx"]:
                    r = requests.get(_standard_url + id + ".cif")
                    content = r.text
                    _assert_valid_file(content, id)
                elif format == "mmtf":
                    r = requests.get(_mmtf_url + id)
                    content = r.content
                    _assert_valid_file(r.text, id)
                elif format == "fasta":
                    r = requests.get(_fasta_url + id)
                    content = r.text
                    _assert_valid_file(content, id)
                else:
                    raise ValueError(f"Format '{format}' is not supported")
                
                if file is None:
                    if format in _binary_formats:
                        file = io.BytesIO(content)
                    else:
                        file = io.StringIO(content)
                else:
                    mode = "wb+" if format in _binary_formats else "w+"
                    with open(file, mode) as f:
                        f.write(content)
        
        files.append(file)
    if verbose:
        print("\nDone")
    # If input was a single ID, return only a single path
    if single_element:
        return files[0]
    else:
        return files


def _assert_valid_file(response_text, pdb_id):
    """
    Checks whether the response is an actual structure file
    or the response a *404* error due to invalid PDB ID.
    """
    # Structure file and FASTA file retrieval
    # have different error messages
    if any(err_msg in response_text for err_msg in [
        "404 Not Found",
        "<title>RCSB Protein Data Bank Error Page</title>",
        "No fasta files were found."
    ]):
        raise RequestError("PDB ID {:} is invalid".format(pdb_id))



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
import tempfile
import numpy as np
from requests.exceptions import ConnectionError
import pytest
import biotite.database.entrez as entrez
import biotite.sequence.io.fasta as fasta
from biotite.database import RequestError
from ..util import cannot_connect_to


NCBI_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/"


@pytest.mark.skipif(
    cannot_connect_to(NCBI_URL),
    reason="NCBI Entrez is not available"
)
@pytest.mark.parametrize(
    "common_name, as_file_like",
    itertools.product([False, True], [False, True])
)
def test_fetch(common_name, as_file_like):
    path = None if as_file_like else tempfile.gettempdir()
    db_name = "Protein" if common_name else "protein"
    file = entrez.fetch(
        "1L2Y_A", path, "fa", db_name, "fasta", overwrite=True
    )
    fasta_file = fasta.FastaFile.read(file)
    prot_seq = fasta.get_sequence(fasta_file)
    assert len(prot_seq) == 20

@pytest.mark.skipif(
    cannot_connect_to(NCBI_URL),
    reason="NCBI Entrez is not available"
)
@pytest.mark.parametrize("as_file_like", [False, True])
def test_fetch_single_file(as_file_like):
    if as_file_like:
        file_name = None
    else:
        file = tempfile.NamedTemporaryFile("r", suffix=".fa")
        file_name = file.name
    
    downloaded_file_name = entrez.fetch_single_file(
        ["1L2Y_A", "3O5R_A"], file_name, "protein", "fasta"
    )
    fasta_file = fasta.FastaFile.read(downloaded_file_name)
    prot_seqs = fasta.get_sequences(fasta_file)
    assert len(prot_seqs) == 2

    if not as_file_like:
        file.close()

@pytest.mark.skipif(
    cannot_connect_to(NCBI_URL),
    reason="NCBI Entrez is not available"
)
def test_fetch_invalid():
    with pytest.raises(RequestError):
        file = entrez.fetch("xxxx", tempfile.gettempdir(), "fa", "protein",
                            "fasta", overwrite=True)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
import pytest
import numpy as np
import biotite.sequence as seq
import biotite.sequence.align as align
from .util import sequences


@pytest.mark.parametrize(
    "gap_penalty, seed, threshold, direction, score_only",
    itertools.product(
        [-10, (-10,-1)],
        [(0, 0), (11, 11), (20, 19), (30, 29)],
        [20, 100, 500],
        ["both", "upstream","downstream"],
        [False, True]
    )
)
def test_simple_alignment(gap_penalty, seed, threshold,
                          direction, score_only):
    """
    Test `align_local_gapped()` by comparing the output to
    `align_optimal()`.
    This test uses a pair of highly similar short sequences.
    """
    # Cyclotide C, Uniprot: P86843
    seq1 = seq.ProteinSequence("gvpcaescvwipctvtallgcsckdkvcyld")
    # Cyclotide F, Uniprot: P86846
    seq2 = seq.ProteinSequence("gipcgescvfipcissvvgcsckskvcyld")
    matrix = align.SubstitutionMatrix.std_protein_matrix()

    ref_alignments = align.align_optimal(
        seq1, seq2, matrix,
        gap_penalty=gap_penalty, local=True
    )
    # Limit reference alignment range to seed
    # if the alignment does not extend in both directions
    for alignment in ref_alignments:
        seed_index = np.where(alignment.trace[:,0] == seed[0])[0][0]
        if direction == "upstream":
            alignment.trace = alignment.trace[:seed_index + 1]
        elif direction == "downstream":
            alignment.trace = alignment.trace[seed_index:]
        alignment.score = align.score(alignment, matrix, gap_penalty)
    
    test_result = align.align_local_gapped(
        seq1, seq2, matrix, seed, threshold, gap_penalty,
        1000, direction, score_only
    )

    if score_only:
        test_score = test_result
        # All optimal alignments have the same score
        assert test_score == ref_alignments[0].score
    else:
        test_alignments = test_result
        assert len(test_alignments) == len(ref_alignments)
        for alignment in test_alignments:
            assert alignment in ref_alignments


@pytest.mark.parametrize(
    "gap_penalty, score_only, seq_indices",
    itertools.product(
        [-10, (-10,-1)],
        [False, True],
        [(i,j) for i in range(10) for j in range(i+1)]
    )
)
def test_complex_alignment(sequences, gap_penalty, score_only,
                           seq_indices):
    """
    Test `align_local_gapped()` by comparing the output to
    `align_optimal()`.
    This test uses a set of long sequences, which are pairwise compared.
    The threshold should be chosen sufficiently large so
    `align_local_gapped()` can return the optimal alignment(s).
    """
    MAX_NUMBER = 100
    # The linear gap penalty for longer gaps easily exceeds
    # a small threshold -> increase threshold for linear penalty
    THRESHOLD = 200 if isinstance(gap_penalty, int) else 50
    
    matrix = align.SubstitutionMatrix.std_protein_matrix()
    index1, index2 = seq_indices
    seq1 = sequences[index1]
    seq2 = sequences[index2]

    ref_alignments = align.align_optimal(
        seq1, seq2, matrix,
        gap_penalty=gap_penalty, local=True, max_number=MAX_NUMBER
    )
    # Select the center of the alignment as seed
    trace = ref_alignments[0].trace
    trace = trace[(trace != -1).all(axis=1)]
    seed = trace[len(trace) // 2]
    
    test_result = align.align_local_gapped(
        seq1, seq2, matrix, seed, THRESHOLD, gap_penalty,
        MAX_NUMBER, "both", score_only
    )

    if score_only:
        test_score = test_result
        # All optimal alignments have the same score
        assert test_score == ref_alignments[0].score
    else:
        try:
            test_alignments = test_result
            assert test_alignments[0].score == ref_alignments[0].score
            # Test if the score is also correctly calculated
            assert align.score(test_alignments[0], matrix, gap_penalty) \
                == ref_alignments[0].score
            if len(ref_alignments) < MAX_NUMBER \
               and len(test_alignments) < MAX_NUMBER:
                    # Only test if the exact same alignments were created,
                    # if the number of traces was not limited by MAX_NUMBER
                    for i, alignment in enumerate(test_alignments):
                        try:
                            assert alignment in ref_alignments
                        except AssertionError:
                            # Edge case:
                            # In rare case the local alignment may be
                            # slightly longer on the upstream side for
                            # 'align_local_ungapped()', since the
                            # upstream side is handled in an inverted
                            # manner
                            # However this does not effect the score
                            # Consequently, the exception is ignored
                            # if the alignment is longer than all
                            # reference alignments
                            if len(alignment) <= max(
                                [len(ali) for ali in ref_alignments]
                            ):
                                raise
        except AssertionError:
            print(f"Missing test alignment at index {i}:")
            print()
            print(test_alignments[i])
            print("\n")
            print("First reference alignment:")
            print()
            print(ref_alignments[0])
            raise


@pytest.mark.parametrize(
    "gap_penalty, direction, score_only, should_raise",
    itertools.product(
        [-10, (-10,-1)],
        ["both", "upstream","downstream"],
        [False, True],
        [False, True]
    )
)
def test_max_table_size(gap_penalty, direction, score_only, should_raise):
    """
    Check if the `max_table_size` parameter in `align_local_gapped()`
    raises the expected `MemoryError` if the aligned regions get too
    large.
    """
    if should_raise:
        # This table size is exceed in this test case...
        max_table_size = 1_000_000
    else:
        # ... and this one is not
        max_table_size = 1_000_000_000

    # Align a long random sequence to itself,
    # effectively resulting in a global alignment 
    np.random.seed(0)
    seq1 = seq.NucleotideSequence()
    seq1.code = np.random.randint(len(seq1.alphabet), size=10000)

    matrix = align.SubstitutionMatrix.std_nucleotide_matrix()
    # Local alignment starts in the center of the sequences
    seed = (len(seq1) // 2, len(seq1) // 2)
    threshold = 100

    if should_raise:
        with pytest.raises(MemoryError):
            align.align_local_gapped(
                seq1, seq1, matrix, seed, threshold, gap_penalty, 1,
                direction, score_only, max_table_size
            )
    else:
        result = align.align_local_gapped(
            seq1, seq1, matrix, seed, threshold, gap_penalty, 1,
            direction, score_only, max_table_size
        )
        if not score_only and direction == "both":
            alignment = result[0]
            # Expect that no gaps are introduced
            assert len(alignment) == len(seq1)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import NamedTemporaryFile
import itertools
import glob
from os.path import join, basename
import numpy as np
import pytest
import biotite.structure as struc
import biotite.structure.io as strucio
import biotite.structure.io.xtc as xtc
import biotite.structure.io.trr as trr
import biotite.structure.io.tng as tng
import biotite.structure.io.dcd as dcd
import biotite.structure.io.netcdf as netcdf
import biotite.structure.io.pdbx as pdbx
from ..util import data_dir, cannot_import


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize("format", ["trr", "xtc", "tng", "dcd", "netcdf"])
def test_array_conversion(format):
    template = strucio.load_structure(
        join(data_dir("structure"), "1l2y.mmtf")
    )[0]
    # Add fake box
    template.box = np.diag([1,2,3])
    if format == "trr":
        traj_file_cls = trr.TRRFile
    if format == "xtc":
        traj_file_cls = xtc.XTCFile
    if format == "tng":
        traj_file_cls = tng.TNGFile
    if format == "dcd":
        traj_file_cls = dcd.DCDFile
    if format == "netcdf":
        traj_file_cls = netcdf.NetCDFFile
    traj_file = traj_file_cls.read(
        join(data_dir("structure"), f"1l2y.{format}")
    )
    ref_array = traj_file.get_structure(template)

    traj_file = traj_file_cls()
    traj_file.set_structure(ref_array)
    temp = NamedTemporaryFile("w+b")
    traj_file.write(temp.name)

    traj_file = traj_file_cls.read(temp.name)
    temp.close()
    array = traj_file.get_structure(template)
    assert ref_array.bonds == array.bonds
    assert ref_array.equal_annotation_categories(array)
    assert ref_array.box == pytest.approx(array.box)
    assert ref_array.coord == pytest.approx(array.coord, abs=1e-2)


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "format, start, stop, step, chunk_size",
    itertools.product(
        ["trr", "xtc", "tng", "dcd", "netcdf"],
        [None, 2],
        [None, 17],
        [None, 2],
        [None, 3]
    )
)
def test_mmtf_consistency(format, start, stop, step, chunk_size):
    if format == "netcdf" and stop is not None and step is not None:
        # Currently, there is an inconsistency in in MDTraj's
        # NetCDFTrajectoryFile class:
        # In this class the number of frames in the output arrays
        # is dependent on the 'stride' parameter
        return
    
    # MMTF is used as reference for consistency check
    # due to higher performance
    ref_traj = strucio.load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    ref_traj = ref_traj[slice(start, stop, step)]
    
    # Template is first model of the reference
    template = ref_traj[0]
    if format == "trr":
        traj_file_cls = trr.TRRFile
    if format == "xtc":
        traj_file_cls = xtc.XTCFile
    if format == "tng":
        traj_file_cls = tng.TNGFile
    if format == "dcd":
        traj_file_cls = dcd.DCDFile
    if format == "netcdf":
        traj_file_cls = netcdf.NetCDFFile
    traj_file = traj_file_cls.read(
        join(data_dir("structure"), f"1l2y.{format}"),
        start, stop, step, chunk_size=chunk_size
    )
    test_traj = traj_file.get_structure(template)
    test_traj_time = traj_file.get_time()
    
    if format not in ["dcd", "netcdf"]:
        # The time starts at 1.0 and increases by 1.0 each step
        # -> can be tested against 'range()' function
        # Shift to ensure time starts at 0
        test_traj_time -= 1
        start = start if start is not None else 0
        stop = stop if stop is not None else 38     # 38 models in 1l2y
        step = step if step is not None else 1
        assert test_traj_time.astype(int).tolist() \
            == list(range(start, stop, step))

    assert test_traj.stack_depth() == ref_traj.stack_depth()
    # 1l2y has no box
    # no assert np.array_equal(test_traj.box, ref_traj.box)
    assert test_traj.bonds == ref_traj.bonds
    assert test_traj.equal_annotation_categories(ref_traj)
    assert test_traj.coord == pytest.approx(ref_traj.coord, abs=1e-2)


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "format, start, stop, step",
    itertools.product(
        ["trr", "xtc", "tng", "dcd", "netcdf"],
        [None, 2],
        [None, 17],
        [None, 2]
    )
)
def test_read_iter(format, start, stop, step):
    """
    Compare aggregated yields of :func:`read_iter()` with the values
    from a corresponding :class:`TrajectoryFile` object.
    """
    if format == "netcdf" and step is not None:
        # Currently, there is an inconsistency in in MDTraj's
        # NetCDFTrajectoryFile class:
        # In this class the number of frames in the output arrays
        # is dependent on the 'stride' parameter
        return
    
    if format == "trr":
        traj_file_cls = trr.TRRFile
    if format == "xtc":
        traj_file_cls = xtc.XTCFile
    if format == "tng":
        traj_file_cls = tng.TNGFile
    if format == "dcd":
        traj_file_cls = dcd.DCDFile
    if format == "netcdf":
        traj_file_cls = netcdf.NetCDFFile
    file_name = join(data_dir("structure"), f"1l2y.{format}")
    
    traj_file = traj_file_cls.read(file_name, start, stop, step)
    ref_coord = traj_file.get_coord()
    ref_box = traj_file.get_box()
    ref_time = traj_file.get_time()

    test_coord = []
    test_box = []
    test_time = []
    for coord, box, time in traj_file.read_iter(file_name, start, stop, step):
        test_coord.append(coord)
        test_box.append(box)
        test_time.append(time)
    # Convert list to NumPy array
    test_coord = np.stack(test_coord)
    test_box = np.stack(test_box)
    test_time = np.stack(test_time)

    assert test_coord.tolist() == ref_coord.tolist()
    
    if ref_box is None:
        assert (test_box == None).all()
    else:
        assert test_box.tolist() == ref_box.tolist()
    
    if ref_time is None:
        assert (test_time == None).all()
    else:
        assert test_time.tolist() == ref_time.tolist()


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "format, start, stop, step",
    itertools.product(
        ["trr", "xtc", "tng", "dcd", "netcdf"],
        [None, 2],
        [None, 17],
        [None, 2]
    )
)
def test_read_iter_structure(format, start, stop, step):
    """
    Compare aggregated yields of :func:`read_iter_structure()` with the
    return value of :func:`get_structure()` from a corresponding
    :class:`TrajectoryFile` object.
    """
    if format == "netcdf" and step is not None:
        # Currently, there is an inconsistency in in MDTraj's
        # NetCDFTrajectoryFile class:
        # In this class the number of frames in the output arrays
        # is dependent on the 'stride' parameter
        return
    
    template = strucio.load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    
    if format == "trr":
        traj_file_cls = trr.TRRFile
    if format == "xtc":
        traj_file_cls = xtc.XTCFile
    if format == "tng":
        traj_file_cls = tng.TNGFile
    if format == "dcd":
        traj_file_cls = dcd.DCDFile
    if format == "netcdf":
        traj_file_cls = netcdf.NetCDFFile
    file_name = join(data_dir("structure"), f"1l2y.{format}")
    
    traj_file = traj_file_cls.read(file_name, start, stop, step)
    ref_traj = traj_file.get_structure(template)
    
    test_traj = struc.stack(
        [frame for frame in traj_file_cls.read_iter_structure(
            file_name, template, start, stop, step
        )]
    )
    
    assert test_traj == ref_traj


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "format, n_models, n_atoms, include_box, include_time",
    itertools.product(
        ["trr", "xtc", "tng", "dcd", "netcdf"],
        [1, 100],
        [1, 1000],
        [False, True],
        [False, True],
    )
)
def test_write_iter(format, n_models, n_atoms, include_box, include_time):
    """
    Expect that `write_iter()` and `write()` create files with equal content.
    """
    if format == "trr":
        traj_file_cls = trr.TRRFile
    if format == "xtc":
        traj_file_cls = xtc.XTCFile
    if format == "tng":
        # TNG files do only write time when more than one frame is
        # written to file; 'write_iter()' writes only one frame per
        # 'write()' call, hence time is not written
        traj_file_cls = tng.TNGFile
        include_time = False
    if format == "dcd":
        traj_file_cls = dcd.DCDFile
        # DCD format does not support simulation time
        include_time = False
    if format == "netcdf":
        traj_file_cls = netcdf.NetCDFFile

    # Generate random coordinate dataset content
    np.random.seed(0)
    coord = np.random.rand(n_models, n_atoms, 3) * 100
    box = np.random.rand(n_models, 3, 3) * 100 if include_box else None
    # time is evenly spaced for TNG compatibility
    time = np.linspace(0, 10, n_models) if include_time else None


    ref_file = NamedTemporaryFile("w+b")
    traj_file = traj_file_cls()
    traj_file.set_coord(coord)
    traj_file.set_box(box)
    traj_file.set_time(time)
    traj_file.write(ref_file.name)
    
    traj_file = traj_file_cls.read(ref_file.name)
    ref_coord = traj_file.get_coord()
    ref_box = traj_file.get_box()
    ref_time = traj_file.get_time()
    ref_file.close()


    test_file = NamedTemporaryFile("w+b")
    traj_file_cls.write_iter(test_file.name, coord, box, time)

    traj_file = traj_file_cls.read(test_file.name)
    test_coord = traj_file.get_coord()
    test_box = traj_file.get_box()
    test_time = traj_file.get_time()
    test_file.close()


    assert np.allclose(test_coord, ref_coord, atol=1e-2)
    if include_box:
        assert np.allclose(test_box, ref_box, atol=1e-2)
    if include_time:
        assert np.allclose(test_time, ref_time, atol=1e-2)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.database.entrez"
__author__ = "Patrick Kunzmann, Maximilian Dombrowsky"
__all__ = ["check_for_errors"]

from ..error import RequestError


# Taken from https://github.com/kblin/ncbi-entrez-error-messages
_error_messages = [
    "Error reading from remote server",
    "Bad gateway",
    "Bad Gateway",
    "Cannot process ID list",
    "server is temporarily unable to service your request",
    "Service unavailable",
    "Server Error",
    "ID list is empty",
    "Resource temporarily unavailable",
    "Failed to retrieve sequence",
    "Failed to understand id",
]


def check_for_errors(message):
    """
    Check for common error messages in NCBI Entrez database responses.
    
    Parameters
    ----------
    message : str
        The message received from NCBI Entrez. 
    
    Raises
    ------
    RequestError
        If the message contains an error message.
    """
    # Error always appear at the end of message
    message_end = message[-100:]
    for error_msg in _error_messages:
        if error_msg in message_end:
            raise RequestError(error_msg)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.info"
__author__ = "Patrick Kunzmann"
__all__ = ["standardize_order"]

import numpy as np
from .atoms import residue
from ..residues import get_residue_starts
from ..error import BadStructureError


_atom_name_cache = {}


def standardize_order(atoms):
    """
    Get an index array for an input :class:`AtomArray` or
    :class:`AtomArrayStack` that reorders the atoms for each residue
    to obtain the standard *RCSB PDB* atom order.

    The standard atom order is determined from the reference residues in
    the official *Chemical Component Dictionary*.
    If a residue of the input structure contains additional atoms that
    are not present in the reference residue, these indices to these
    atoms are appended to the end of the respective residue.
    A example for this are optional hydrogen atoms, that appear due to
    protonation.

    Parameters
    ----------
    atoms : AtomArray, shape=(n,) or AtomArrayStack, shape=(m,n)
        Input structure with atoms that are potentially not in the
        *standard* order.
    
    Returns
    -------
    indices : ndarray, dtype=int, shape=(n,)
        When this index array is applied on the input `atoms`,
        the atoms for each residue are reordered to obtain the
        standard *RCSB PDB* atom order.
    
    Raises
    ------
    BadStructureError
        If the input `atoms` have duplicate atoms (same atom name)
        within a residue.
    
    Examples
    --------

    Use as single residue as example.

    >>> residue = atom_array[atom_array.res_id == 1]
    >>> print(residue)
        A       1  ASN N      N        -8.901    4.127   -0.555
        A       1  ASN CA     C        -8.608    3.135   -1.618
        A       1  ASN C      C        -7.117    2.964   -1.897
        A       1  ASN O      O        -6.634    1.849   -1.758
        A       1  ASN CB     C        -9.437    3.396   -2.889
        A       1  ASN CG     C       -10.915    3.130   -2.611
        A       1  ASN OD1    O       -11.269    2.700   -1.524
        A       1  ASN ND2    N       -11.806    3.406   -3.543
        A       1  ASN H1     H        -8.330    3.957    0.261
        A       1  ASN H2     H        -8.740    5.068   -0.889
        A       1  ASN H3     H        -9.877    4.041   -0.293
        A       1  ASN HA     H        -8.930    2.162   -1.239
        A       1  ASN HB2    H        -9.310    4.417   -3.193
        A       1  ASN HB3    H        -9.108    2.719   -3.679
        A       1  ASN HD21   H       -11.572    3.791   -4.444
        A       1  ASN HD22   H       -12.757    3.183   -3.294

    Reverse the atom array.
    Consequently, this also changes the atom order within the residue.

    >>> reordered = residue[np.arange(len(residue))[::-1]]
    >>> print(reordered)
        A       1  ASN HD22   H       -12.757    3.183   -3.294
        A       1  ASN HD21   H       -11.572    3.791   -4.444
        A       1  ASN HB3    H        -9.108    2.719   -3.679
        A       1  ASN HB2    H        -9.310    4.417   -3.193
        A       1  ASN HA     H        -8.930    2.162   -1.239
        A       1  ASN H3     H        -9.877    4.041   -0.293
        A       1  ASN H2     H        -8.740    5.068   -0.889
        A       1  ASN H1     H        -8.330    3.957    0.261
        A       1  ASN ND2    N       -11.806    3.406   -3.543
        A       1  ASN OD1    O       -11.269    2.700   -1.524
        A       1  ASN CG     C       -10.915    3.130   -2.611
        A       1  ASN CB     C        -9.437    3.396   -2.889
        A       1  ASN O      O        -6.634    1.849   -1.758
        A       1  ASN C      C        -7.117    2.964   -1.897
        A       1  ASN CA     C        -8.608    3.135   -1.618
        A       1  ASN N      N        -8.901    4.127   -0.555

    The order is restored with the exception of the N-terminus protonation.

    >>> restored = reordered[info.standardize_order(reordered)]
    >>> print(restored)
        A       1  ASN N      N        -8.901    4.127   -0.555
        A       1  ASN CA     C        -8.608    3.135   -1.618
        A       1  ASN C      C        -7.117    2.964   -1.897
        A       1  ASN O      O        -6.634    1.849   -1.758
        A       1  ASN CB     C        -9.437    3.396   -2.889
        A       1  ASN CG     C       -10.915    3.130   -2.611
        A       1  ASN OD1    O       -11.269    2.700   -1.524
        A       1  ASN ND2    N       -11.806    3.406   -3.543
        A       1  ASN H2     H        -8.740    5.068   -0.889
        A       1  ASN HA     H        -8.930    2.162   -1.239
        A       1  ASN HB2    H        -9.310    4.417   -3.193
        A       1  ASN HB3    H        -9.108    2.719   -3.679
        A       1  ASN HD21   H       -11.572    3.791   -4.444
        A       1  ASN HD22   H       -12.757    3.183   -3.294
        A       1  ASN H3     H        -9.877    4.041   -0.293
        A       1  ASN H1     H        -8.330    3.957    0.261
    """
    reordered_indices = np.zeros(atoms.array_length(), dtype=int)

    starts = get_residue_starts(atoms, add_exclusive_stop=True)
    for i in range(len(starts)-1):
        start = starts[i]
        stop = starts[i+1]

        res_name = atoms.res_name[start]
        standard_atom_names = _atom_name_cache.get(res_name)
        if standard_atom_names is None:
            standard_atom_names = residue(res_name).atom_name
            _atom_name_cache[res_name] = standard_atom_names
        
        reordered_indices[start : stop] = _reorder(
            atoms.atom_name[start : stop], standard_atom_names
        ) + start

    return reordered_indices


def _reorder(origin, target):
    """
    Create indices to `origin`, that changes the order of `origin`,
    so that the order is the same as in `target`.

    Indices for elements of `target` that are not in `origin`
    are ignored.
    Indices for elements of `origin` that are not in `target`
    are appended to the end of the returned array.


    Parameters
    ----------
    origin : ndarray, dtype=str
        The atom names to reorder.
    target : ndarray, dtype=str
        The atom names in target order.
    
    Returns
    -------
    indices : ndarray, dtype=int
        Indices for `origin` that that changes the order of `origin`
        to the order of `target`.
    """
    target_hits, origin_hits = np.where(
        target[:, np.newaxis] == origin[np.newaxis, :]
    )

    counts = np.bincount(target_hits, minlength=len(target))
    if (counts > 1).any():
        counts = np.bincount(target_hits, minlength=len(target))
        # Identify which atom is duplicate
        duplicate_i = np.where(
            counts > 1
        )[0][0]
        duplicate_name = target[duplicate_i]
        raise BadStructureError(
            f"Input structure has duplicate atom '{duplicate_name}'"
        )

    if len(origin_hits) < len(origin):
        # The origin structure has additional atoms
        # to the target structure
        # -> Identify which atoms are missing in the target structure
        # and append these to the end of the residue
        missing_atom_mask = np.bincount(
            origin_hits, minlength=len(origin)
        ).astype(bool)
        return np.concatenate([
            origin_hits,
            np.where(~missing_atom_mask)[0]
        ])
    else:
        return origin_hits


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.mmtf"
__author__ = "Patrick Kunzmann"
__all__ = ["MMTFFile"]

import io
from collections.abc import MutableMapping
import struct
import copy
import numpy as np
import msgpack
from ....file import File, is_binary
from ...error import BadStructureError
from .decode import decode_array
from .encode import encode_array


class MMTFFile(File, MutableMapping):
    """
    This class represents a MMTF file.
    
    When reading a file, the *MessagePack* unpacker is used to create
    a dictionary of the file content.
    This dictionary is accessed by indexing the :class:`MMTFFile`
    instance directly with the dictionary keys.
    If the dictionary value is an encoded array, the value automatically
    decoded.
    Decoded arrays are always returned as :class:`ndarray` instances.
    
    Examples
    --------
    
    >>> import os.path
    >>> mmtf_file = MMTFFile.read(os.path.join(path_to_structures, "1l2y.mmtf"))
    >>> print(mmtf_file["title"])
    NMR Structure of Trp-Cage Miniprotein Construct TC5b
    >>> print(mmtf_file["chainNameList"])
    ['A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A'
     'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A' 'A'
     'A' 'A']
    """
    
    def __init__(self):
        super().__init__()
        self._content = {}
        self._content["mmtfVersion"] = "1.0.0"
        self._content["mmtfProducer"] = "UNKNOWN"
    
    @classmethod
    def read(self, file):
        """
        Read a MMTF file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        
        Returns
        -------
        file_object : MMTFFile
            The parsed file.
        """
        mmtf_file = MMTFFile()
        # File name
        if isinstance(file, str):
            with open(file, "rb") as f:
                mmtf_file._content = msgpack.unpackb(
                    f.read(), use_list=True, raw=False
                )
        # File object
        else:
            if not is_binary(file):
                raise TypeError("A file opened in 'binary' mode is required")
            mmtf_file._content = msgpack.unpackb(
                file.read(), use_list=True, raw=False
            )
        return mmtf_file
    
    def write(self, file):
        """
        Write contents into a MMTF file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be written to.
            Alternatively, a file path can be supplied.
        """
        packed_bytes = msgpack.packb(
            self._content, use_bin_type=True, default=_encode_numpy
        )
        if isinstance(file, str):
            with open(file, "wb") as f:
                f.write(packed_bytes)
        else:
            if not is_binary(file):
                raise TypeError("A file opened in 'binary' mode is required")
            file.write(packed_bytes)
    
    def __copy_fill__(self, clone):
        super().__copy_fill__(clone)
        clone._content = copy.deepcopy(self._content)
    
    def get_codec(self, key):
        """
        Obtain the codec ID of an MMTF encoded value.
        
        Parameters
        ----------
        key : str
            The key for the potentially encoded value.
        
        Returns
        -------
        codec : int or None
            The codec ID. `None` if the value is not encoded.
        """
        data = self._content[key]
        if isinstance(data, bytes) and data[0] == 0:
            codec = struct.unpack(">i", data[0:4])[0]
            return codec
        else:
            return None
    
    def get_length(self, key):
        """
        Obtain the length of an MMTF encoded value.
        
        Parameters
        ----------
        key : str
            The key for the potentially encoded value.
        
        Returns
        -------
        codec : int or None
            The length of the `bytes` array.
            `None` if the value is not encoded.
        """
        data = self._content[key]
        if isinstance(data, bytes) and data[0] == 0:
            length = struct.unpack(">i", data[4:8])[0]
            return length
        else:
            return None
    
    def get_param(self, key):
        """
        Obtain the parameter of an MMTF encoded value.
        
        Parameters
        ----------
        key : str
            The key for the potentially encoded value.
        
        Returns
        -------
        codec : int or None
            The parameter of the encoded value.
            `None` if the value is not encoded.
        """
        data = self._content[key]
        if isinstance(data, bytes) and data[0] == 0:
            param = struct.unpack(">i", data[8:12])[0]
            return param
        else:
            return None
    
    def set_array(self, key, array, codec, param=0):
        length = len(array)
        raw_bytes = encode_array(array, codec, param)
        data = struct.pack(">i", codec) \
             + struct.pack(">i", length) \
             + struct.pack(">i", param) \
             + raw_bytes
        self._content[key] = data
    
    def __getitem__(self, key):
        data = self._content[key]
        if isinstance(data, bytes) and data[0] == 0:
            # MMTF specific format -> requires decoding
            codec     = struct.unpack(">i", data[0:4 ])[0]
            length    = struct.unpack(">i", data[4:8 ])[0]
            param     = struct.unpack(">i", data[8:12])[0]
            raw_bytes = data[12:]
            return decode_array(codec, raw_bytes, param)
        else:
            return data
    
    def __setitem__(self, key, item):
        if isinstance(item, np.ndarray):
            raise TypeError("Arrays that need to be encoded must be addeed "
                            "via 'set_array()'")
        self._content[key] = item
    
    def __delitem__(self, key):
        del self._content[key]
    
    def __iter__(self):
        return self._content.__iter__()
    
    def __len__(self):
        return len(self._content)
    
    def __contains__(self, item):
        return item in self._content


def _encode_numpy(item):
    """
    Convert NumPy scalar types to native Python types,
    as *Msgpack* cannot handle NumPy types (e.g. float32).

    The function is given to the Msgpack packer as value for the
    `default` parameter.
    """
    if isinstance(item, np.generic):
        return item.item()
    else:
        raise TypeError(f"can not serialize '{type(item).__name__}' object")



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence"
__author__ = "Patrick Kunzmann"
__all__ = ["CodonTable"]

import copy
from os.path import join, dirname, realpath
import numpy as np
from numbers import Integral
from .seqtypes import NucleotideSequence, ProteinSequence


# Abbreviations
_NUC_ALPH = NucleotideSequence.alphabet_unamb
_PROT_ALPH = ProteinSequence.alphabet

# Multiplier array that converts a codon in code representation
# into a unique integer
_radix = len(_NUC_ALPH)
_radix_multiplier = np.array([_radix**n for n in (2,1,0)], dtype=int)


class CodonTable(object):
    """
    A :class:`CodonTable` maps a codon (sequence of 3 nucleotides) to an
    amino acid.
    It also defines start codons. A :class:`CodonTable`
    takes/outputs either the symbols or code of the codon/amino acid.
    
    Furthermore, this class is able to give a list of codons that
    corresponds to a given amino acid.
    
    The :func:`load()` method allows loading of NCBI codon tables.
    
    Objects of this class are immutable.
    
    Parameters
    ----------
    codon_dict : dict of (str -> str)
        A dictionary that maps codons to amino acids. The keys must be
        strings of length 3 and the values strings of length 1
        (all upper case).
        The dictionary must provide entries for all 64 possible codons.
    starts : iterable object of str
        The start codons. Each entry must be a string of length 3
        (all upper case).
    
    Examples
    --------
    
    Get the amino acid coded by a given codon (symbol and code):
        
    >>> table = CodonTable.default_table()
    >>> print(table["ATG"])
    M
    >>> print(table[(1,2,3)])
    14
        
    Get the codons coding for a given amino acid (symbol and code):
        
    >>> table = CodonTable.default_table()
    >>> print(table["M"])
    ('ATG',)
    >>> print(table[14])
    ((0, 2, 0), (0, 2, 2), (1, 2, 0), (1, 2, 1), (1, 2, 2), (1, 2, 3))
    """
    
    # For efficient mapping of codon codes to amino acid codes,
    # especially in in the 'map_codon_codes()' function, the class
    # maps each possible codon into a unique number using a radix based
    # approach.
    # For example the codon (3,1,2) would be represented as
    # 3*16 + 1*4 + 2**1 = 53

    # file for builtin codon tables from NCBI
    _table_file = join(dirname(realpath(__file__)), "codon_tables.txt")
    
    def __init__(self, codon_dict, starts):
        # Check if 'starts' is iterable object of length 3 string
        for start in starts:
            if not isinstance(start, str) or len(start) != 3:
                raise ValueError(f"Invalid codon '{start}' as start codon")
        # Internally store codons as single unique numbers
        start_codon_codes = np.array(
            [_NUC_ALPH.encode_multiple(start) for start in starts], dtype=int
        )
        self._starts = CodonTable._to_number(start_codon_codes)
        # Use -1 as error code
        # The array uses the number representation of codons as index
        # and stores the corresponding symbol codes for amino acids
        self._codons = np.full(_radix**3, -1, dtype=int)
        for key, value in codon_dict.items():
            codon_code = _NUC_ALPH.encode_multiple(key)
            codon_number = CodonTable._to_number(codon_code)
            aa_code = _PROT_ALPH.encode(value)
            self._codons[codon_number] = aa_code
        if (self._codons == -1).any():
            # Find the missing codon
            missing_index = np.where(self._codons == -1)[0][0]
            codon_code = CodonTable._to_codon(missing_index) 
            codon = _NUC_ALPH.decode_multiple(codon_code)
            codon_str = "".join(codon)
            raise ValueError(
                f"Codon dictionary does not contain codon '{codon_str}'"
            )

    def __repr__(self):
        """Represent CodonTable as a string for debugging."""
        return f"CodonTable({self.codon_dict()}, {self.start_codons()})"

    def __eq__(self, item):
        if not isinstance(item, CodonTable):
            return False
        if self.codon_dict() != item.codon_dict():
            return False
        if self.start_codons() != item.start_codons():
            return False
        return True

    def __ne__(self, item):
        return not self == item

    def __getitem__(self, item):
        if isinstance(item, str):
            if len(item) == 1:
                # Amino acid -> return possible codons
                aa_code = _PROT_ALPH.encode(item)
                codon_numbers = np.where(self._codons == aa_code)[0]
                codon_codes = CodonTable._to_codon(codon_numbers)
                codons = tuple(
                    ["".join(_NUC_ALPH.decode_multiple(codon_code))
                    for codon_code in codon_codes]
                )
                return codons
            elif len(item) == 3:
                # Codon -> return corresponding amino acid
                codon_code = _NUC_ALPH.encode_multiple(item)
                codon_number = CodonTable._to_number(codon_code)
                aa_code = self._codons[codon_number]
                aa = _PROT_ALPH.decode(aa_code)
                return aa
            else:
                raise ValueError(f"'{item}' is an invalid index")
        elif isinstance(item, int):
            # Code for amino acid -> return possible codon codes
            codon_numbers = np.where(self._codons == item)[0]
            codon_codes = tuple(CodonTable._to_codon(codon_numbers))
            codon_codes = tuple([tuple(code) for code in codon_codes])
            return codon_codes
        else:
            # Code for codon as any iterable object
            # Code for codon -> return corresponding amino acid codes
            if len(item) != 3:
                raise ValueError(
                    f"{item} is an invalid sequence code for a codon"
                )
            codon_number = CodonTable._to_number(item)
            aa_code = self._codons[codon_number]
            return aa_code
    
    def map_codon_codes(self, codon_codes):
        """
        Efficiently map multiple codons to the corresponding amino
        acids.
        
        Parameters
        ----------
        codon_codes : ndarray, dtype=int, shape=(n,3)
            The codons to be translated into amino acids.
            The codons are given as symbol codes.
            *n* is the amount of codons.
        
        Returns
        -------
        aa_codes : ndarray, dtype=int, shape=(n,)
            The amino acids as symbol codes.
        
        Examples
        --------
        >>> dna = NucleotideSequence("ATGGTTTAA")
        >>> sequence_code = dna.code
        >>> print(sequence_code)
        [0 3 2 2 3 3 3 0 0]
        >>> # Reshape to get codons
        >>> codon_codes = sequence_code.reshape(-1, 3)
        >>> print(codon_codes)
        [[0 3 2]
         [2 3 3]
         [3 0 0]]
        >>> # Map to amino acids
        >>> aa_codes = CodonTable.default_table().map_codon_codes(codon_codes)
        >>> print(aa_codes)
        [10 17 23]
        >>> # Put into a protein sequence
        >>> protein = ProteinSequence()
        >>> protein.code = aa_codes
        >>> print(protein)
        MV*
        """
        if codon_codes.shape[-1] != 3:
            raise ValueError(
                f"Codons must be length 3, "
                f"but size of last dimension is {codon_codes.shape[-1]}"
            )
        codon_numbers = CodonTable._to_number(codon_codes)
        aa_codes = self._codons[codon_numbers]
        return aa_codes
        
    def codon_dict(self, code=False):
        """
        Get the codon to amino acid mappings dictionary.
        
        Parameters
        ----------
        code : bool
            If true, the dictionary contains keys and values as code.
            Otherwise, the dictionary contains strings for codons and
            amino acid. (Default: False)
        
        Returns
        -------
        codon_dict : dict
            The dictionary mapping codons to amino acids.
        """
        if code:
            return {tuple(CodonTable._to_codon(codon_number)): aa_code
                    for codon_number, aa_code in enumerate(self._codons)}
        else:
            return {"".join(_NUC_ALPH.decode_multiple(codon_code)):
                        _PROT_ALPH.decode(aa_code)
                    for codon_code, aa_code
                    in self.codon_dict(code=True).items()}
    
    def is_start_codon(self, codon_codes):
        codon_numbers = CodonTable._to_number(codon_codes)
        return np.isin(codon_numbers, self._starts)
    
    def start_codons(self, code=False):
        """
        Get the start codons of the codon table.
        
        Parameters
        ----------
        code : bool
            If true, the code will be returned instead of strings.
            (Default: False)
        
        Returns
        -------
        start_codons : tuple
            The start codons. Contains strings or tuples, depending on
            the `code` parameter.
        """
        if code:
            return tuple(
                [tuple(CodonTable._to_codon(codon_number))
                 for codon_number in self._starts]
            )
        else:
            return tuple(
                ["".join(_NUC_ALPH.decode_multiple(codon_code))
                 for codon_code in self.start_codons(code=True)]
            )
    
    def with_start_codons(self, starts):
        """
        Create an new :class:`CodonTable` with the same codon mappings,
        but changed start codons.
        
        Parameters
        ----------
        starts : iterable object of str
            The new start codons.
        
        Returns
        -------
        new_table : CodonTable
            The codon table with the new start codons.
        """
        # Copy this table and replace the start codons
        new_table = copy.deepcopy(self)
        start_codon_codes = np.array(
            [_NUC_ALPH.encode_multiple(start) for start in starts], dtype=int
        )
        new_table._starts = CodonTable._to_number(start_codon_codes)
        return new_table
    
    def with_codon_mappings(self, codon_dict):
        """
        Create an new :class:`CodonTable` with partially changed codon
        mappings.
        
        Parameters
        ----------
        codon_dict : dict of (str -> str)
            The changed codon mappings.
        
        Returns
        -------
        new_table : CodonTable
            The codon table with changed codon mappings.
        """
        # Copy this table and replace the codon
        new_table = copy.deepcopy(self)
        for key, value in codon_dict.items():
            codon_code = _NUC_ALPH.encode_multiple(key)
            codon_number = CodonTable._to_number(codon_code)
            aa_code = _PROT_ALPH.encode(value)
            new_table._codons[codon_number] = aa_code
        return new_table

    def __str__(self):
        string = ""
        # ['A', 'C', 'G', 'T']
        bases = _NUC_ALPH.get_symbols()
        for b1 in bases:
            for b2 in bases:
                for b3 in bases:
                    codon = b1 + b2 + b3
                    string += codon + " " + self[codon]
                    # Indicator for start codon
                    codon_code = _NUC_ALPH.encode_multiple(codon)
                    if CodonTable._to_number(codon_code) in self._starts:
                        string += " i "
                    else:
                        string += "   "
                    # Add space for next codon
                    string += " "*3
                # Remove terminal space
                string = string [:-6]
                # Jump to next line
                string += "\n"
            # Add empty line
            string += "\n"
        # Remove the two terminal new lines
        string = string[:-2]
        return string

    @staticmethod
    def _to_number(codons):
        if not isinstance(codons, np.ndarray):
            codons = np.array(list(codons), dtype=int)
        return np.sum(_radix_multiplier * codons, axis=-1)

    @staticmethod
    def _to_codon(numbers):
        if isinstance(numbers, Integral):
            # Only a single number
            return CodonTable._to_codon(np.array([numbers]))[0]
        if not isinstance(numbers, np.ndarray):
            numbers = np.array(list(numbers), dtype=int)
        codons = np.zeros(numbers.shape + (3,), dtype=int)
        for n in (2,1,0):
            val = _radix**n
            digit = numbers // val
            codons[..., -(n+1)] = digit
            numbers = numbers - digit * val
        return codons

    @staticmethod
    def load(table_name):
        """
        Load a NCBI codon table.
        
        Parameters
        ----------
        table_name : str or int
            If a string is given, it is interpreted as official NCBI
            codon table name (e.g. "Vertebrate Mitochondrial").
            An integer is interpreted as NCBI codon table ID.
        
        Returns
        -------
        table : CodonTable
            The NCBI codon table.
        """
        # Loads codon tables from codon_tables.txt
        with open(CodonTable._table_file, "r") as f:
            lines = f.read().split("\n")
        
        # Extract data for codon table from file
        table_found = False
        aa = None
        init = None
        base1 = None
        base2 = None
        base3 = None
        for line in lines:
            if not line:
                table_found = False
            if type(table_name) == int and line.startswith("id"):
                # remove identifier 'id'
                if table_name == int(line[2:]):
                    table_found = True
            elif type(table_name) == str and line.startswith("name"):
                # Get list of table names from lines
                # (separated with ';')
                # remove identifier 'name'
                names = [name.strip() for name in line[4:].split(";")]
                if table_name in names:
                    table_found = True
            if table_found:
                if line.startswith("AA"):
                    #Remove identifier
                    aa = line[5:].strip()
                elif line.startswith("Init"):
                    init = line[5:].strip()
                elif line.startswith("Base1"):
                    base1 = line[5:].strip()
                elif line.startswith("Base2"):
                    base2 = line[5:].strip()
                elif line.startswith("Base3"):
                    base3 = line[5:].strip()
        
        # Create codon table from data
        if aa is not None and init is not None \
            and base1 is not None and base2 is not None and base3 is not None:
                symbol_dict = {}
                starts = []
                # aa, init and baseX all have the same length
                for i in range(len(aa)):
                    codon = base1[i] + base2[i] + base3[i]
                    if init[i] == "i":
                        starts.append(codon)
                    symbol_dict[codon] = aa[i]
                return CodonTable(symbol_dict, starts)
        else:
            raise ValueError(f"Codon table '{table_name}' was not found")

    @staticmethod
    def table_names():
        """
        The possible codon table names for :func:`load()`.
        
        Returns
        -------
        names : list of str
            List of valid codon table names.
        """
        with open(CodonTable._table_file, "r") as f:
            lines = f.read().split("\n")
        names = []
        for line in lines:
            if line.startswith("name"):
                names.extend([name.strip() for name in line[4:].split(";")])
        return names
    
    @staticmethod
    def default_table():
        """
        The default codon table.
        The table is equal to the NCBI "Standard" codon table,
        with the difference that only "ATG" is a start codon.
        
        Returns
        -------
        table : CodonTable
            The default codon table.
        """
        return _default_table


_default_table = CodonTable.load("Standard").with_start_codons(["ATG"])



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.io.fastq"
__author__ = "Patrick Kunzmann"

from collections import OrderedDict
from ...sequence import Sequence
from ...alphabet import AlphabetError, LetterAlphabet
from ...seqtypes import NucleotideSequence
from ...align.alignment import Alignment

__all__ = ["get_sequence", "get_sequences", "set_sequence", "set_sequences"]


def get_sequence(fastq_file, header=None):
    """
    Get a sequence and quality scores from a `FastqFile` instance.
    
    Parameters
    ----------
    fastq_file : FastqFile
        The `FastqFile` to be accessed.
    header : str, optional
        The identifier to get the sequence and scores from.
        By default, the first sequence of the file is returned.
    
    Returns
    -------
    sequence : NucleotideSequence
        The requested sequence.
    scores : ndarray, dtype=int
        The requested scores.
    """
    if header is not None:
        seq_str, scores = fastq_file[header]
    else:
        # Return first (and probably only) sequence of file
        seq_str = None
        scores = None
        for seq_str, scores in fastq_file.values():
            break
        if seq_str is None:
            raise ValueError("File does not contain any sequences")
    processed_seq_str = seq_str.replace("U","T").replace("X","N")
    return NucleotideSequence(processed_seq_str), scores


def get_sequences(fastq_file):
    """
    Get a dictionary from a `FastqFile` instance,
    where identifiers are keys and sequence-score-tuples are values.
    
    Parameters
    ----------
    fastq_file : FastqFile
        The `Fastqile` to be accessed.
    
    Returns
    -------
    seq_dict : dict
        A dictionary containing identifiers as keys and
        (`NucleotideSequence`, `ndarray`) tuples as values.
    """
    seq_dict = OrderedDict()
    for header, (seq_str, scores) in fastq_file.items():
        processed_seq_str = seq_str.replace("U","T").replace("X","N")
        seq_dict[header] = NucleotideSequence(processed_seq_str), scores
    return seq_dict


def set_sequence(fastq_file, sequence, scores, header=None, as_rna=False):
    """
    Set a sequence and a quality score array in a `FastqFile` instance.
    
    Parameters
    ----------
    fastq_file : FastqFile
        The `FastqFile` to be accessed.
    sequence : NucleotideSequence
        The sequence to be set.
    scores : ndarray, dtype=int
        The quality scores to be set.
    header : str, optional
        The identifier for the sequence. Default is 'sequence'.
    as_rna : bool, optional
        If set to true, the sequence symbol ``'T'`` will be replaced
        by ``'U'``.
    """
    if header is None:
        header = "sequence"
    fastq_file[header] = _convert_to_string(sequence, as_rna), scores


def set_sequences(fastq_file, sequence_dict, as_rna=False):
    """
    Set sequences in a `FastqFile` instance from a dictionary.
    
    Parameters
    ----------
    fastq_file : FastqFile
        The `FastqFile` to be accessed.
    sequence_dict : dict
        A dictionary containing the sequences and scores to be set.
        Identifiers are keys,
        (`NucleotideSequence`, `ndarray`) tuples are values.
    as_rna : bool, optional
        If set to true, the sequence symbol ``'T'`` will be replaced
        by ``'U'``.
    """
    for header, (sequence, scores) in sequence_dict.items():
        fastq_file[header] = _convert_to_string(sequence, as_rna), scores


def _convert_to_string(sequence, as_rna):
    if as_rna:
        return(str(sequence).replace("T", "U"))
    else:
        return(str(sequence))


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import TemporaryFile
import glob
import itertools
from os.path import join, splitext
import numpy as np
import pytest
from pytest import approx
import biotite
import biotite.structure as struc
import biotite.structure.info as info
import biotite.structure.io.mmtf as mmtf
import biotite.structure.io.pdbx as pdbx
from ..util import data_dir


def test_get_model_count():
    mmtf_file = mmtf.MMTFFile.read(join(data_dir("structure"), "1l2y.mmtf"))
    test_model_count = mmtf.get_model_count(mmtf_file)
    ref_model_count = mmtf.get_structure(mmtf_file).stack_depth()
    assert test_model_count == ref_model_count


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "*.mmtf"))
)
def test_codecs(path):
    mmtf_file = mmtf.MMTFFile.read(path)
    for key in mmtf_file:
        if mmtf_file.get_codec(key) is not None:
            codec = mmtf_file.get_codec(key)
            param = mmtf_file.get_param(key)
            array1 = mmtf_file[key]
            mmtf_file.set_array(key, array1, codec, param)
            array2 = mmtf_file[key]
            if array1.dtype == np.float32:
                if param != 0:
                    tol = 1/param
                else:
                    tol = 0
                assert np.isclose(array1, array2, atol=tol).all()
            else:
                assert (array1 == array2).all()


@pytest.mark.parametrize(
    "path, model",
    itertools.product(
        glob.glob(join(data_dir("structure"), "*.mmtf")),
        [None, 1, -1]
    )
)
def test_array_conversion(path, model):
    mmtf_file = mmtf.MMTFFile.read(path)
    try:
        a1 = mmtf.get_structure(mmtf_file, model=model, include_bonds=True)
    except biotite.InvalidFileError:
        if model is None:
            # The file cannot be parsed into an AtomArrayStack,
            # as the models contain different numbers of atoms
            # -> skip this test case
            return
        else:
            raise
    
    mmtf_file = mmtf.MMTFFile()
    mmtf.set_structure(mmtf_file, a1)
    temp = TemporaryFile("w+b")
    mmtf_file.write(temp)

    temp.seek(0)
    mmtf_file = mmtf.MMTFFile.read(temp)
    temp.close()
    a2 = mmtf.get_structure(mmtf_file, model=model, include_bonds=True)
    
    for category in a1.get_annotation_categories():
        assert a1.get_annotation(category).tolist() == \
               a2.get_annotation(category).tolist()
    assert a1.coord.flatten().tolist() == \
           approx(a2.coord.flatten().tolist(), abs=1e-3)
    assert a1.bonds == a2.bonds
    if a1.box is not None:
        assert np.allclose(a1.box, a2.box)


@pytest.mark.parametrize(
    "path, model",
    itertools.product(
        glob.glob(join(data_dir("structure"), "*.mmtf")),
        [None, 1, -1]
    )
)
def test_pdbx_consistency(path, model):
    cif_path = splitext(path)[0] + ".cif"
    mmtf_file = mmtf.MMTFFile.read(path)
    try:
        a1 = mmtf.get_structure(mmtf_file, model=model)
    except biotite.InvalidFileError:
        if model is None:
            # The file cannot be parsed into an AtomArrayStack,
            # as the models contain different numbers of atoms
            # -> skip this test case
            return
        else:
            raise
    
    pdbx_file = pdbx.PDBxFile.read(cif_path)
    a2 = pdbx.get_structure(pdbx_file, model=model)
    
    # Sometimes mmCIF files can have 'cell' entry
    # but corresponding MMTF file has not 'unitCell' entry
    # -> Do not assert for dummy entry in mmCIF file
    # (all vector elements = {0, 1})
    if a2.box is not None and not ((a2.box == 0) | (a2.box == 1)).all():
        assert np.allclose(a1.box, a2.box)
    # MMTF might assign some residues, that PDBx assigns as 'hetero',
    # as 'non-hetero' if they are RNA/DNA or peptide linking
    try:
        assert a1.hetero.tolist() == \
               a2.hetero.tolist()
    except AssertionError:
        conflict_residues = np.unique(
            a1.res_name[a1.hetero != a2.hetero]
        )
        for res in conflict_residues:
            assert info.link_type(res) in [
                "L-PEPTIDE LINKING", "PEPTIDE LINKING",
                "DNA LINKING", "RNA LINKING"
            ]
    # Test the remaining categories
    for category in [
        c for c in a1.get_annotation_categories() if c != "hetero"
    ]:
        assert a1.get_annotation(category).tolist() == \
               a2.get_annotation(category).tolist()
    assert a1.coord.flatten().tolist() == \
           approx(a2.coord.flatten().tolist(), abs=1e-3)


def test_extra_fields():
    path = join(data_dir("structure"), "1l2y.mmtf")
    mmtf_file = mmtf.MMTFFile.read(path)
    stack1 = mmtf.get_structure(
        mmtf_file,
        extra_fields=[
            "atom_id", "b_factor", "occupancy", "charge"
        ]
    )

    mmtf_file == mmtf.MMTFFile()
    mmtf.set_structure(mmtf_file, stack1)
    
    stack2 = mmtf.get_structure(
        mmtf_file,
        extra_fields=[
            "atom_id", "b_factor", "occupancy", "charge"
        ]
    )
    
    assert stack1.atom_id.tolist() == stack2.atom_id.tolist()
    assert stack1.b_factor.tolist() == approx(stack2.b_factor.tolist())
    assert stack1.occupancy.tolist() == approx(stack2.occupancy.tolist())
    assert stack1.charge.tolist() == stack2.charge.tolist()


def test_numpy_objects():
    """
    Test whether the Msgpack encoder is able to handle NumPy values
    (e.g. np.float32) properly.

    Only check if no error occurs.
    """
    mmtf_file = mmtf.MMTFFile()
    mmtf_file["A float"] = np.float32(42.0)
    mmtf_file["A list"] = [np.int64(1), np.int64(2), np.int64(3)]
    mmtf_file["A dictionary"] = {"a": np.int64(1), "b": np.int64(2)}
    temp = TemporaryFile("w+b")
    mmtf_file.write(temp)
    temp.close()



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import numpy as np
import pytest
import biotite.sequence as seq
import biotite.sequence.align as align
from .util import sequences



def test_alignment_str():
    """
    Test reading alignments from string.
    """
    seq1 = seq.NucleotideSequence("ACCTGA")
    seq2 = seq.NucleotideSequence("TATGCT")
    ali_str = ["A-CCTGA----",
               "----T-ATGCT"]
    trace = align.Alignment.trace_from_strings(ali_str)
    alignment = align.Alignment([seq1, seq2], trace, None)
    assert str(alignment).split("\n") == ali_str

def test_conversion_to_symbols():
    """
    Test conversion of alignments to strings.
    """
    seq_str1 = "HAKLPRDD--WKL--"
    seq_str2 = "HA--PRDDADWKLHH"
    seq_str3 = "HA----DDADWKLHH"
    seq_strings = [seq_str1, seq_str2, seq_str3]
    sequences = [seq.ProteinSequence(seq_str.replace("-",""))
                 for seq_str in seq_strings]
    trace = align.Alignment.trace_from_strings(seq_strings)
    alignment = align.Alignment(sequences, trace, score=None)
    # Test the conversion bach to strings of symbols
    symbols = align.get_symbols(alignment)
    symbols = ["".join([sym if sym is not None else "-" for sym in sym_list])
               for sym_list in symbols]
    assert symbols == seq_strings

def test_identity():
    """
    Test correct calculation of `get_sequence_identity()` via a known
    test case.
    """
    seq_str1 = "--HAKLPRDD--WL--"
    seq_str2 = "FRHA--QRTDADWLHH"
    seq_strings = [seq_str1, seq_str2]
    sequences = [seq.ProteinSequence(seq_str.replace("-",""))
                 for seq_str in seq_strings]
    trace = align.Alignment.trace_from_strings(seq_strings)
    alignment = align.Alignment(sequences, trace, score=None)
    # Assert correct sequence identity calculation
    modes = ["all", "not_terminal", "shortest"]
    values = [6/16, 6/12, 6/10]
    for mode, value in zip(modes, values):
        assert align.get_sequence_identity(alignment, mode=mode) == value

@pytest.mark.parametrize("mode", ["all", "not_terminal", "shortest"])
def test_pairwise_identity(sequences, mode):
    """
    Test correct calculation of `get_pairwise_sequence_identity()` via
    pairwise calls of `get_sequence_identity()`.
    """
    sequences = sequences
    msa, _, _, _ = align.align_multiple(
        sequences,
        matrix=align.SubstitutionMatrix.std_protein_matrix()
    )
    
    ref_identity_matrix = np.zeros((len(sequences), len(sequences)))
    for i in range(len(sequences)):
        for j in range(len(sequences)):
            ref_identity_matrix[i,j] = align.get_sequence_identity(
                msa[:, [i,j]], mode=mode
            )
    
    test_identity_matrix = align.get_pairwise_sequence_identity(msa, mode=mode)
    
    # Identity of two equal sequences should be 1, if only the length of
    # the sequence is counted
    if mode == "shortest":
        assert (np.diag(test_identity_matrix) == 1).all()
    # Identity must be between 0 and 1
    assert ((test_identity_matrix <= 1) & (test_identity_matrix >= 0)).all()
    # Identity matrix is symmetric
    assert (test_identity_matrix == test_identity_matrix.T).all()
    # Pairwise identity must be equal in the two functions
    assert (test_identity_matrix == ref_identity_matrix).all()